#!/usr/bin/env python

import os
import re
import tempfile
from unittest import TestCase, main

import deepdiff
import networkx as nx

import neuroarch.nxtools as nxtools

class TestNXTools(TestCase):
    def setUp(self):
        self.g_digraph = nx.DiGraph()
        self.g_digraph.add_nodes_from([(0, {'x': 3}),
                                       (1, {'x': 4}),
                                       (2, {'y': 3}),
                                       (3, {'x': 3})])
        self.g_digraph.add_edges_from([(0, 1, {'a': 1}),
                                       (0, 2, {'b': 1}),
                                       (0, 3, {'a': 1}),
                                       (1, 2, {'b': 2})])

        self.g_digraph_str = nx.DiGraph()
        self.g_digraph_str.add_nodes_from([(0, {'x': 'foo'}),
                                           (1, {'x': 'fooo'}),
                                           (2, {'y': 'bar'}),
                                           (3, {'x': 'bar'})])
        self.g_digraph_str.add_edges_from([(0, 1, {'a': 'qux'}),
                                           (0, 2, {'b': 'mux'}),
                                           (0, 3, {'a': 'qux'}),
                                           (1, 2, {'b': 'mux'})])

        self.g_multi = nx.MultiDiGraph()
        self.g_multi.add_nodes_from([(0, {'x': 3}),
                                     (1, {'x': 4}),
                                     (2, {'y': 3}),
                                     (3, {'x': 3})])
        self.g_multi.add_edges_from([(0, 1, {'a': 1}),
                                     (0, 1, {'a': 1}),
                                     (0, 2, {'b': 1}),
                                     (0, 3, {'a': 1}),
                                     (1, 2, {'b': 2})])

        self.g_multi_str = nx.MultiDiGraph()
        self.g_multi_str.add_nodes_from([(0, {'x': 'foo'}),
                                         (1, {'x': 'fooo'}),
                                         (2, {'y': 'bar'}),
                                         (3, {'x': 'bar'})])
        self.g_multi_str.add_edges_from([(0, 1, {'a': 'qux'}),
                                         (0, 1, {'a': 'qux'}),
                                         (0, 2, {'b': 'mux'}),
                                         (0, 3, {'a': 'qux'}),
                                         (1, 2, {'b': 'mux'})])

    def test_nodes_has(self):
        result = nxtools.nodes_has(self.g_digraph, [1, 2], 'x', 4)
        self.assertItemsEqual(result, [1])

    def test_nodes_has_regex(self):
        result = nxtools.nodes_has(self.g_digraph_str, [1, 2], 'x',
                                   'foo.*', re.search)
        self.assertItemsEqual(result, [1])
        
    def test_edges_has_digraph(self):
        result = nxtools.edges_has(self.g_digraph, [(0, 1), (0, 2)], 'a', 1)
        self.assertItemsEqual(result, [(0, 1)])

    def test_edges_has_digraph_regex(self):
        result = nxtools.edges_has(self.g_digraph_str, [(0, 1), (0, 2)], 'a',
                                   'qu.*', re.search)
        self.assertItemsEqual(result, [(0, 1)])

    def test_edges_has_multi(self):
        result = nxtools.edges_has(self.g_multi, [(0, 1, 0), (0, 2, 0)], 'a', 1)
        self.assertItemsEqual(result, [(0, 1, 0)])

    def test_edges_has_multi_regex(self):
        result = nxtools.edges_has(self.g_multi_str, [(0, 1, 0), (0, 2, 0)], 
                                   'a', 'qu.*', re.search)
        self.assertItemsEqual(result, [(0, 1, 0)])

    def test_all_nodes_has(self):
        result = nxtools.all_nodes_has(self.g_digraph, 'x', 3)
        self.assertItemsEqual(result, [0, 3])

    def test_all_nodes_has_regex(self):
        result = nxtools.all_nodes_has(self.g_digraph_str, 'x', 'foo.*', re.search)
        self.assertItemsEqual(result, [0, 1])

    def test_all_edges_has_digraph(self):
        result = nxtools.all_edges_has(self.g_digraph, 'a', 1)
        self.assertItemsEqual(result, [(0, 1), (0, 3)])

    def test_all_edges_has_digraph_regex(self):
        result = nxtools.all_edges_has(self.g_digraph_str, 'a', 'qu.*', re.search)
        self.assertItemsEqual(result, [(0, 1), (0, 3)])

    def test_all_edges_has_multi(self):
        result = nxtools.all_edges_has(self.g_multi, 'a', 1)
        self.assertItemsEqual(result,
                             [(0, 1, 0), (0, 1, 1), (0, 3, 0)])

    def test_all_edges_has_multi_regex(self):
        result = nxtools.all_edges_has(self.g_multi_str, 'a',
                                       'qu.*', re.search)
        self.assertItemsEqual(result,
                             [(0, 1, 0), (0, 1, 1), (0, 3, 0)])

    def test_out_nodes_has(self):
        result = nxtools.out_nodes_has(self.g_digraph, [0], 'x', 3)
        self.assertItemsEqual(result, [3])

    def test_out_nodes_has_regex(self):
        result = nxtools.out_nodes_has(self.g_digraph_str, [0], 'x',
                                       'foo.*', re.search)
        self.assertItemsEqual(result, [1])

    def test_in_nodes_has(self):
        result = nxtools.in_nodes_has(self.g_digraph, [1], 'x', 3)
        self.assertItemsEqual(result, [0])

    def test_in_nodes_has_regex(self):
        result = nxtools.in_nodes_has(self.g_digraph_str, [2], 'x',
                                      '.*ooo', re.search)
        self.assertItemsEqual(result, [1])

    def test_out_edges_has_digraph(self):
        result = nxtools.out_edges_has(self.g_digraph, [0], 'b', 1)
        self.assertItemsEqual(result, [(0, 2)])

    def test_out_edges_has_digraph_regex(self):
        result = nxtools.out_edges_has(self.g_digraph_str, [0],
                                       'b', 'mu.*', re.search)
        self.assertItemsEqual(result, [(0, 2)])

    def test_out_edges_has_multi(self):
        result = nxtools.out_edges_has(self.g_multi, [0], 'b', 1)
        self.assertItemsEqual(result, [(0, 2, 0)])

    def test_out_edges_has_multi_regex(self):
        result = nxtools.out_edges_has(self.g_multi_str, [0], 
                                       'b', 'mu.*', re.search)
        self.assertItemsEqual(result, [(0, 2, 0)])

    def test_in_edges_has_digraph(self):
        result = nxtools.in_edges_has(self.g_digraph, [2], 'b', 1)
        self.assertItemsEqual(result, [(0, 2)])

    def test_in_edges_has_digraph_regex(self):
        result = nxtools.in_edges_has(self.g_digraph_str, [1, 2],
                                      'a', 'qu.*', re.search)
        self.assertItemsEqual(result, [(0, 1)])

    def test_in_edges_has_multi(self):
        result = nxtools.in_edges_has(self.g_multi, [2], 'b', 1)
        self.assertItemsEqual(result, [(0, 2, 0)])

    def test_in_edges_has_multi_regex(self):
        result = nxtools.in_edges_has(self.g_multi_str, [1],
                                      'a', 'qu.*', re.search)
        self.assertItemsEqual(result, [(0, 1, 0), (0, 1, 1)])
                             
    def test_find_nonmatching_dict_pairs(self):
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}], [{1: 2}]) == ([], [])        
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}, {3: 4}],
                                           [{1: 2}, {3: 4}]) == ([], [])
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}, {3: 4}],
                                           [{1: 2}, {3: 5}]) == ([{3: 4}], [{3: 5}])
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}, {1: 2}, {3: 4}],
                                  [{1: 2}, {1: 2}, {3: 5}]) == ([{3: 4}], [{3: 5}])
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}, {1: 2}, {3: 4}, {3: 4}],
                                           [{1: 2}, {1: 2}, {3: 5}, {3: 5}]) == \
            ([{3: 4}, {3: 4}], [{3: 5}, {3: 5}])            
        assert nxtools.find_nonmatching_dict_pairs([{1: 2}, {1: 2}, {3: 4}, {3: 5}],
                                  [{1: 2}, {1: 2}, {3: 5}, {3: 5}]) == \
            ([{3: 4}], [{3: 5}])

    def test_is_isomorphic_attr(self):
        g0 = nx.MultiDiGraph()
        g0.add_node(0, **{'name': 'foo',
                                  'node_type': 'neuron'})
        g0.add_node(1, **{'name': 'bar',
                                  'node_type': 'neuron'})
        g0.add_node(2, **{'name': 'foo-bar',
                                  'node_type': 'synapse'})
        g0.add_edge(0, 2, **{'edge_type': 'data'})
        g0.add_edge(2, 1, **{'edge_type': 'data'})

        g1 = nx.MultiDiGraph()
        g1.add_node('a', **{'name': 'foo',
                                    'node_type': 'neuron'})
        g1.add_node('b', **{'name': 'bar',
                                    'node_type': 'neuron'})
        g1.add_node('c', **{'name': 'foo-bar',
                                    'node_type': 'synapse'})
        g1.add_edge('a', 'c', **{'edge_type': 'data'})
        g1.add_edge('c', 'b', **{'edge_type': 'data'})

        g2 = nx.MultiDiGraph()
        g2.add_node('a', **{'name': 'foo',
                                    'node_type': 'neuron'})
        g2.add_node('b', **{'name': 'bar',
                                    'node_type': 'neuron'})
        g2.add_node('c', **{'name': 'foo-bar',
                                    'node_type': 'synapse'})
        g2.add_edge('a', 'c', **{'edge_type': 'xxxx'})
        g2.add_edge('c', 'b', **{'edge_type': 'data'})

        assert nxtools.is_isomorphic_attr(g0, g1)
        assert not nxtools.is_isomorphic_attr(g0, g2)

    def test_iso_attr_diff_multidigraph_no_multiedges(self):
        g0 = nx.MultiDiGraph()
        g0.add_node(0, **{'name': 'foo',
                                  'node_type': 'neuron'})
        g0.add_node(1, **{'name': 'bar',
                                  'node_type': 'neuron'})
        g0.add_node(2, **{'name': 'foo-bar',
                                  'node_type': 'synapse'})
        g0.add_edge(0, 2, **{'edge_type': 'data'})
        g0.add_edge(2, 1, **{'edge_type': 'data'})

        g1 = nx.MultiDiGraph()
        g1.add_node('a', **{'name': 'foo',
                                    'node_type': 'neuron'})
        g1.add_node('b', **{'name': 'bar',
                                    'node_type': 'neuron'})
        g1.add_node('c', **{'name': 'foo-bar',
                                    'node_type': 'synapse'})
        g1.add_edge('a', 'c', **{'edge_type': 'data'})
        g1.add_edge('c', 'b', **{'edge_type': 'data'})

        node_diff, edge_diff = nxtools.iso_attr_diff(g0, g1)
        self.assertDictEqual(node_diff, {})
        self.assertDictEqual(edge_diff, {})

    def test_iso_attr_diff_multidigraph_multiedges(self):
        g0 = nx.MultiDiGraph()
        g0.add_node(0, **{'name': 'foo',
                                  'node_type': 'neuron'})
        g0.add_node(1, **{'name': 'bar',
                                  'node_type': 'neuron'})
        g0.add_node(2, **{'name': 'foo-bar',
                                  'node_type': 'synapse'})
        g0.add_edge(0, 2, **{'edge_type': 'data'})
        g0.add_edge(0, 2, **{'edge_type': 'data'})
        g0.add_edge(2, 1, **{'edge_type': 'data'})

        g1 = nx.MultiDiGraph()
        g1.add_node('a', **{'name': 'foo',
                                    'node_type': 'neuron'})
        g1.add_node('b', **{'name': 'bar',
                                    'node_type': 'neuron'})
        g1.add_node('c', **{'name': 'foo-bar',
                                    'node_type': 'synapse'})
        g1.add_edge('a', 'c', **{'edge_type': 'data'})
        g1.add_edge('a', 'c', **{'edge_type': 'data'})
        g1.add_edge('c', 'b', **{'edge_type': 'data'})

        g2 = nx.MultiDiGraph()
        g2.add_node('a', **{'name': 'foo',
                                    'node_type': 'neuron'})
        g2.add_node('b', **{'name': 'bar',
                                    'node_type': 'neuron'})
        g2.add_node('c', **{'name': 'foo-bar',
                                    'node_type': 'synapse'})
        g2.add_edge('a', 'c', **{'edge_type': 'data'})
        g2.add_edge('a', 'c', **{'edge_type': 'xxxx'})
        g2.add_edge('c', 'b', **{'edge_type': 'data'})

        node_diff, edge_diff = nxtools.iso_attr_diff(g0, g1)
        self.assertDictEqual(node_diff, {})
        self.assertDictEqual(edge_diff, {})

        node_diff, edge_diff = nxtools.iso_attr_diff(g0, g2)

        self.assertDictEqual(node_diff, {})
        self.assertDictEqual(edge_diff, 
            {((0, 2, 0), ('a', 'c', 1)): 
             {'values_changed': ["root['edge_type']: 'data' ===> 'xxxx'"]}})

    def _attr_match(self, a, b):
        return False if deepdiff.DeepDiff(a, b) else True

    def test_read_gexf_digraph(self):
        f = tempfile.mktemp()
        g_orig = nx.DiGraph()
        g_orig.add_nodes_from(['a', 'b'])
        g_orig.add_edge('a', 'b', **{'x': 1})
        nx.write_gexf(g_orig, f)
        g_new = nxtools.read_gexf(f)
        os.unlink(f)

        self.assertTrue(nx.isomorphism.is_isomorphic(g_orig, g_new, 
                                                     self._attr_match, 
                                                     self._attr_match))

    def test_read_gexf_multidigraph(self):
        f = tempfile.mktemp()
        g_orig = nx.MultiDiGraph()
        g_orig.add_nodes_from(['a', 'b'])
        g_orig.add_edge('a', 'b', **{'x': 1})
        g_orig.add_edge('a', 'b', **{'x': 2})
        nx.write_gexf(g_orig, f)
        g_new = nxtools.read_gexf(f)
        os.unlink(f)

        self.assertTrue(nx.isomorphism.is_isomorphic(g_orig, g_new, 
                                                     self._attr_match, 
                                                     self._attr_match))

if __name__ == '__main__':
    main()



#!/usr/bin/env python

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import pickle
import copy
import logging
import sys

import networkx as nx
from pyorient.ogm import Graph, Config

from neuroarch.models import *
from neuroarch.utils import byteify, chunks, get_cluster_ids

import csv

import pandas as pd
import json

from pyorient.serializations import OrientSerialization

def load_swc(file_name):
    """
    Load an SWC file into a DataFrame.
    """

    df = pd.read_csv(file_name, delimiter=' ', header=None, comment='#',
                     names=['sample', 'identifier', 'x', 'y', 'z', 'r', 'parent'],
                     skipinitialspace=True)
    return df

class NTHULoader(object):
    neuropils = {'AL':('AL',['right antennal lobe','antennal lobe','al_r','al','right al']) ,
           'al':('al',['left antennal lobe','antennal lobe','al_l','al','left al']),
           'AMMC':('AMMC',['right antennal mechanosensory and motor center','antennal mechanosensory and motor center','ammc_r','ammc']),
           'ammc':('ammc',['left antennal mechanosensory and motor center','antennal mechanosensory and motor center','ammc_l','ammc']),
           #'CAL',
           #'cal',
           'CCP':('CCP',['right caudalcentral protocerebrum','caudalcentral protocerebrum','ccp_r','ccp','right ccp']),
           'ccp':('ccp',['left caudalcentral protocerebrum','caudalcentral protocerebrum','ccp_l','ccp','left ccp']),
           'CMP':('CMP',['right caudalmedial protocerebrum','caudalmedial protocerebrum','cmp_r','right cmp', 'cmp']),
           'cmp':('cmp',['left caudalmedial protocerebrum','caudalmedial protocerebrum','cmp_l','left cmp', 'cmp']),
           'CVLP':('CVLP',['right caudal ventrolateral protocerebrum','caudal ventrolateral protocerebrum','cvlp_r','right cvlp','cvlp']),
           'cvlp':('cvlp',['left caudal ventrolateral protocerebrum','caudal ventrolateral protocerebrum','cvlp_l','left cvlp','cvlp']),
           'DLP':('DLP',['right dorsolateral protocerebrum','dorsolateral protocerebrum','right dlp','dlp_r','dlp']),
           'dlp':('dlp',['left dorsolateral protocerebrum','dorsolateral protocerebrum','left dlp','dlp_l','dlp']),
           'DMP':('DMP',['right dorsomedial protocerebrum','dorsomedial protocerebrum','right dmp','dmp_r','dmp','icl','right icl','icl_r']),
           'dmp':('dmp',['left dorsomedial protocerebrum','dorsomedial protocerebrum','left dmp','dmp_l','dmp','icl','left icl','icl_l']),
           'EB':('EB',['ellipsoid body','left ellipsoid body','right ellipsoid body','eb_l','eb_r','eb']),
           #'eb',
           'FSPP':('FSPP',['right frontal superpeduncular protocerebrum','frontal superpeduncular protocerebrum','right fspp','fspp_r','fspp']),
           'fspp':('fspp',['left frontal superpeduncular protocerebrum','frontal superpeduncular protocerebrum','left fspp','fspp_l','fspp']),
           'IDFP':('IDFP',['right inferior dorsofrontal protocerebrum','inferior dorsofrontal protocerebrum','right idfp','idfp_r','idfp']),
           'idfp':('idfp',['left inferior dorsofrontal protocerebrum','inferior dorsofrontal protocerebrum','left idfp','idfp_l','idfp']),
           'IDLP':('IDLP',['right inferior dorsolateral protocerebrum','inferior dorsolateral protocerebrum','right idlp','idlp_r','idlp']),
           'idlp':('idlp',['left inferior dorsolateral protocerebrum','inferior dorsolateral protocerebrum','left idlp','idlp_l','idlp']),
           'LAT':('LAT',['right lat','lat_r','lat']),
           'lat':('lat',['left lat','lat_l','lat']),
           'LH':('LH',['right lateral horn','lateral horn','right lh','lh_r','lh']),
           'lh':('lh',['left lateral horn','lateral horn','left lh','lh_l','lh']),
           'LOB':('LOB',['right lobula','lobula','right lob','lob_r','lob','right lo','lo_r','lo']),
           'lob':('lob',['left lobula','lobula','left lob','lob_l','lob','left lo','lo_l','lo']),
           'LOP':('LOP',['right lobula plate','lobula plate','right lop','lop_r','lop']),
           'lop':('lop',['left lobula plate','lobula plate','left lop','lop_l','lop']),
           'MB':('MB',['right mushroom body','mushroom body','right mb','mb_r','mb']),
           'mb':('mb',['left mushroom body','mushroom body','left mb','mb_l','mb']),
           'MED':('MED',['right medulla','medulla','right med','med_r','med','right me','me_r','me']),
           'LAM':('LAM',['right lamina','lamina','right lam','lam_r','lam','right la','la_r','la']),
           'med':('med',['left medulla','medulla','left med','med_l','med','left me','me_l','me']),
           'NOD':('NOD',['right noduli','noduli','right nod','nod_r','nod','right no','no_r','no']),
           'nod':('nod',['left noduli','noduli','left nod','nod_l','nod','left no','no_l','no']),
           'OG':('OG',['right optic glomerulus','optic glomerulus','right og','og_r','og']),
           'og':('og',['right optic glomerulus','optic glomerulus','left og','og_l','og']),
           'OPTU':('OPTU',['right optic tubercle','optic tubercle','right optu','optu_r','optu']),
           'optu':('optu',['left optic tubercle','optic tubercle','left optu','optu_l','optu']),
           'PAN':('PAN',['right proximal antennal protocerebrum','proximal antennal protocerebrum','right pan','pan_r','pan']),
           'pan':('pan',['left proximal antennal protocerebrum','proximal antennal protocerebrum','left pan','pan_l','pan']),
           'PB':('PB',['right protocerebral bridge','protocerebral bridge','right pb','pb_r','pb','pcb','pcb_r','right pcb','left protocerebral bridge','left pb','pb_l','pcb_l','left pcb']),
           #'pb',
           'SDFP':('SDFP',['right superior dorsofrontal protocerebrum','superior dorsofrontal protocerebrum','right sdfp','sdfp_r','sdfp']),
           'sdfp':('sdfp',['left superior dorsofrontal protocerebrum','superior dorsofrontal protocerebrum','left sdfp','sdfp_l','sdfp']),
           'SOG':('SOG',['right subesophageal ganglion','subesophageal ganglion','right sog','sog_r','sog']),
           'sog':('sog',['left subesophageal ganglion','subesophageal ganglion','left sog','sog_l','sog']),
           'SPP':('SPP',['right superpeduncular protocerebrum','superpeduncular protocerebrum','right spp','spp_r','spp']),
           'spp':('spp',['left superpeduncular protocerebrum','superpeduncular protocerebrum','left spp','spp_l','spp']),
           'FB':('FB',['right fanshaped body','fanshaped body','right fb','fb_r','fb','left fanshaped body','left fb','fb_l']),
           #'fb',
           'VLP':('VLP',['right ventrolateral protocerebrum','ventrolateral protocerebrum','right vlp','vlp_r','vlp']),
           'vlp':('vlp',['left ventrolateral protocerebrum','ventrolateral protocerebrum','left vlp','vlp_l','vlp']),
           'VMP':('VMP',['right ventromedial protocerebrum','ventromedial protocerebrum','right vmp','vmp_r','vmp']),
           'unclear':('unclear',['unclear']),
           'vmp':('vmp',['left ventromedial protocerebrum','ventromedial protocerebrum','left vmp','vmp_l','vmp'])}
    
    neurotransmitter_map = {'Cha': 'acetylcholine',
                            'Gad': 'gaba',
                            'Tdc': 'octopamine',
                            'TH': 'dopamine',
                            'Trh': 'serotonin',
                            'VGlut': 'glutamate'}
                            
    def __init__(self, g_orient):
        self.logger = logging.getLogger('vl')
        self.g_orient = g_orient

        # Make sure OrientDB classes exist:
        #self.g_orient.create_all(Node.registry)
        #self.g_orient.create_all(Relationship.registry)

        # Get cluster IDs:
        self.cluster_ids = get_cluster_ids(self.g_orient.client)
        
    def load_neurons(self, file_name,morph_dir):
        ds_fc = self.g_orient.DataSources.query(name='FlyCircuit').first()
        if not ds_fc:
            ds_fc = self.g_orient.DataSources.create(name='FlyCircuit')
        
        with open(file_name, 'rb') as csvfile:
            reader = csv.reader(csvfile, delimiter=';',)
            i = -1
            for neuron in reader:
                i+=1
                print(i)
                # Process a neuron
                # Name Dendrites Axons  Total  Neuropil Locality
                #  0      1        2      3       4        5

                #if neuron[4]=='unclear': continue
                # Check if neuropil exists
                npl = self.g_orient.Neuropils.query(name=NTHULoader.neuropils[neuron[4]][0]).first()
                if not npl:
                    npl = self.g_orient.Neuropils.create(\
                                            name=NTHULoader.neuropils[neuron[4]][0],
                                            synonyms=NTHULoader.neuropils[neuron[4]][1])
                    self.logger.info('created node: {0}({1})'.format(npl.element_type, npl.name))

                locality = True if neuron[5]=='LN' else False
                # Create Neuron Node
                n = self.g_orient.Neurons.create(name=neuron[0], locality=locality)
                self.logger.info('created node: {0}({1})'.format(n.element_type, n.name))
                
                # Create Neurotransmitter Node if required
                nt = None
                neurotransmitter = []
                for key in NTHULoader.neurotransmitter_map:
                    if neuron[0].startswith(key):
                        neurotransmitter.append(NTHULoader.neurotransmitter_map[key])
                if neurotransmitter:
                    nt = self.g_orient.NeurotransmitterDatas.create(name=neuron[0], Transmitters=neurotransmitter)
                    self.logger.info('created node: {0}({1})'.format(nt.element_type, nt.name))
                    
                # Create Arborization Node
                dendrites = {c.split(':')[0]:int(c.split(':')[1]) for c in neuron[1].split(',')}
                axons = {c.split(':')[0]:int(c.split(':')[1]) for c in neuron[2].split(',')}
                arb = self.g_orient.ArborizationDatas.create(name=neuron[0], dendrites=dendrites, axons=axons)
                self.logger.info('created node: {0}({1})'.format(arb.element_type, arb.name))
                
                # Create Morphology Node
                df = load_swc('%s/%s.swc' % (morph_dir, neuron[0]))
                content = byteify(json.loads(df.to_json()))
                content = {}
                content['x'] = df['x'].tolist()
                content['y'] = df['y'].tolist()
                content['z'] = df['z'].tolist()
                content['r'] = df['r'].tolist()
                content['parent'] = df['parent'].tolist()
                content['identifier'] = df['identifier'].tolist()
                content['sample'] = df['sample'].tolist()
            
            
                content.update({'name': neuron[0]})
    
                nm = self.g_orient.client.record_create(self.cluster_ids['MorphologyData'][0],
                                                       {'@morphologydata': content})
                nm = self.g_orient.get_element(nm._rid)
            
                
                # Add content to new node:
                self.g_orient.client.command('update %s content %s' % \
                                         (nm._id, json.dumps(content)))

                
                self.logger.info('created node: {0}({1})'.format(nm.element_type, nm.name))
                
                # Connect nodes
                self.g_orient.Owns.create(npl, n)
                self.g_orient.HasData.create(n, arb)
                self.g_orient.HasData.create(n, nm)
                if nt:
                    self.g_orient.HasData.create(n, nt)
                    self.g_orient.Owns.create(ds_fc, n)
                self.g_orient.Owns.create(ds_fc, nm)
                self.g_orient.Owns.create(ds_fc, arb)

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout,
                        format='%(asctime)s %(name)s %(levelname)s %(message)s')
    g_orient = Graph(Config.from_url('/na_server','root', 'root', initial_drop=False,
                                     serialization_type=OrientSerialization.Binary))# set to True to erase the database
    g_orient.create_all(Node.registry)
    g_orient.create_all(Relationship.registry)

    
    vl = NTHULoader(g_orient)

    vl.load_neurons('all_female_neurons.txt','flycircuit1.1')



#!/usr/bin/env python

"""
NetworkX tools.
"""

# Copyright (c) 2015-2016, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import copy
import itertools
import operator
import re

import deepdiff
import networkx as nx
import numpy as np

def nodes_has(g, nbunch, attr, value, op=operator.eq, flags=0):
    """
    Filter nodes in list by specific attribute value and comparison operator.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    nbunch : non-string iterable
        List of node identifiers.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected nodes.
    """

    assert np.iterable(nbunch) and not isinstance(nbunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    return [n for n in nbunch if attr in g.node[n] \
                and _op(g.node[n][attr], value)]

def all_nodes_has(g, attr, value, op=operator.eq, flags=0):
    """
    Find all nodes in graph with specific attribute value via specified comparison operator.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`.  The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected nodes.
    """

    return nodes_has(g, g.nodes(), attr, value, op, flags)

def edges_has(g, ebunch, attr, value, op=operator.eq, flags=0):
    """
    Filter edges in list with specific attribute value and comparison operator.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    ebunch : non-string iterable
        Edge endpoint tuples. If `g` is a MultiDiGraph, the tuples must also contain
        the keys for each respective edge.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected edge endpoint tuples.
    """

    assert np.iterable(ebunch) and not isinstance(ebunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    if isinstance(g, nx.MultiDiGraph):
        return [(i, j, k) for (i, j, k) in ebunch \
                if attr in g.edge[i][j][k] and \
                _op(g.edge[i][j][k][attr], value)]
    else:
        return [(i, j) for (i, j) in ebunch \
                if attr in g.edge[i][j] and \
                _op(g.edge[i][j][attr], value)]

def all_edges_has(g, attr, value, op=operator.eq, flags=0):
    """
    Find all edges in graph with specific attribute value via specified comparison operator.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected edge endpoint tuples.
    """

    if isinstance(g, nx.MultiDiGraph):
        e = g.edges(keys=True)
    else:
        e = g.edges()
    return edges_has(g, e, attr, value, op, flags)

def out_nodes_has(g, nbunch, attr, value, op=operator.eq, flags=0):
    """
    Find outgoing nodes with specific attribute value.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    n : non-string iterable
        List of node identifiers.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected node identifiers.
    """

    assert np.iterable(nbunch) and not isinstance(nbunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    result = set()
    for n in nbunch:
        result.update([n for n in g.successors(n) \
                       if attr in g.node[n] and _op(g.node[n][attr], value)])
    return list(result)

def in_nodes_has(g, nbunch, attr, value, op=operator.eq, flags=0):
    """
    Find incoming nodes with specific attribute value.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    nbunch : object
        List of node identifiers.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of selected node identifiers.
    """

    assert np.iterable(nbunch) and not isinstance(nbunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    result = set()
    for n in nbunch:
        result.update([n for n in g.predecessors(n) \
                       if attr in g.node[n] and _op(g.node[n][attr], value)])
    return list(result)

def out_edges_has(g, nbunch, attr, value, op=operator.eq, flags=0):
    """
    Find outgoing edges with specific attribute value.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    nbunch : iterable
        List of node identifiers.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of edge endpoint tuples. If `g` is a MultiDiGraph,
        the tuples contain the endpoints and key for each respective edge.
    """

    assert np.iterable(nbunch) and not isinstance(nbunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    result = set()
    for n in nbunch:
        if isinstance(g, nx.MultiDiGraph):
            result.update([(i, j, k) for (i, j, k, data) in \
                           g.out_edges([n], data=True, keys=True) \
                           if attr in data and _op(data[attr], value)])
        else:
            result.update([(i, j) for (i, j, data) in \
                           g.out_edges([n], data=True) \
                           if attr in data and _op(data[attr], value)])
    return list(result)

def in_edges_has(g, nbunch, attr, value, op=operator.eq, flags=0):
    """
    Find incoming edges with specific attribute value.

    Parameters
    ----------
    g : networkx.Graph
        NetworkX graph.
    nbunch : iterable
        List of node identifiers.
    attr : str
        Attribute name.
    value : object
        Attribute value.
    op : function
        Comparison operator; default is `operator.eq`. The first operand
        is the node attribute and the second is `value` if the operator
        is not `re.search` or `re.match`; otherwise, `value` is assumed to be
        the regular expression pattern used for matching.
    flags : int
        Flags to pass to regular expression matcher. Ignored if 
        `op` is not `re.search` or `re.match`.

    Returns
    -------
    result : list
        List of edge endpoint tuples. If `g` is a MultiDiGraph,
        the tuples contain the endpoints and key for each respective edge.
    """

    assert np.iterable(nbunch) and not isinstance(nbunch, str)
    if op in [re.search, re.match]:
        _op = lambda s, p: op(p, s, flags)
    else:
        _op = op
    result = set()
    for n in nbunch:
        if isinstance(g, nx.MultiDiGraph):
            result.update([(i, j, k) for (i, j, k, data) in \
                           g.in_edges([n], data=True, keys=True) \
                           if attr in data and _op(data[attr], value)])
        else:
            result.update([(i, j) for (i, j, data) in \
                           g.in_edges([n], data=True) \
                           if attr in data and _op(data[attr], value)])
    return list(result)

def find_nonmatching_dict_pairs(a, b):
    """
    Find nonmatching pairs of dicts.

    Given two sets of dicts, try to match each dict in the first set
    with an identical dict in the second dict. Return any remaining pairs of
    non-matching dicts.

    Parameters
    ----------
    a, b : iterable
        Sets of dicts to analyze. The sets must contain the same number of dicts.

    Returns
    -------
    x, y : iterable
        Sets of dicts in `a` and `b`, respectively, that do not match.
    """

    assert len(a) == len(b)
    a = copy.deepcopy(a)
    b = copy.deepcopy(b)
    while True:
        # Look for matches between remaining dicts:
        pairs = itertools.product(a, b)
        for i, j in pairs:

            # Remove matching pair:
            if not deepdiff.DeepDiff(i, j):
                a.remove(i)
                b.remove(j)
                break

        # If no more matches are found, we are done:
        else:
            break
    return a, b

def is_isomorphic_attr(g0, g1):
    """
    Check whether two property graphs are isomorphic.

    Returns True if the graphs are structurally isomorphic and the attributes
    of each node and edge in one graph are identical to those in the isomorphic
    nodes and edges in the other graph.

    Parameters
    ----------
    g0, g1 : networkx.Graph
        NetworkX graphs. Graphs must be of the same type.
    
    Returns
    -------
    result : bool
        True if the graphs are isomorphic.
    """
    
    d = lambda a, b: False if deepdiff.DeepDiff(a, b) else True
    return nx.isomorphism.is_isomorphic(g0, g1,
                                        node_match=d,
                                        edge_match=d)

def iso_attr_diff(g0, g1):
    """
    Find differences in node/edge attributes of structurally isomorphic graphs.

    Parameters
    ----------
    g0, g1 : networkx.Graph
        NetworkX graphs. Graphs must be of the same type.

    Returns
    -------
    node_diff : dict
        Each key is a tuple containing the isomorphic node IDs in `g0` and `g1`,
        respectively; the value contains the corresponding attribute differences.
    edge_diff : dict
        Each key is a tuple containing the isomorphic edge ID tuples for `g0`
        and `g1`, respectively. If the graphs are MultiDiGraph instances, the edge
        ID tuples each contain three values: (from, to, key). Each value of 
        `edge_diff` contains the corresponding attribute differences.
    """

    assert type(g0) == type(g1)
    node_diff = {}
    edge_diff = {}
    if isinstance(g0, nx.MultiDiGraph):
        matcher = nx.isomorphism.MultiDiGraphMatcher(g0, g1)
        if not matcher.is_isomorphic():
            raise ValueError('graphs are not structurally isomorphic')
        for i0, i1 in matcher.mapping.items():
            d = deepdiff.DeepDiff(g0.node[i0], g1.node[i1])
            if d:
                node_diff[(i0, i1)] = d
        for i0, j0 in g0.edges():
            i1 = matcher.mapping[i0]
            j1 = matcher.mapping[j0]

            # Find pairs of edge attributes that don't match:
            d0_nomatch, d1_nomatch = \
                find_nonmatching_dict_pairs(g0.edge[i0][j0].values(), 
                                            g1.edge[i1][j1].values())

            # Find the keys that correspond to the nonmatching pairs of
            # attributes:
            d0_all = copy.deepcopy(g0.edge[i0][j0])
            d1_all = copy.deepcopy(g1.edge[i1][j1])
            for d0, d1 in zip(d0_nomatch, d1_nomatch):

                # Look for key corresponding to attribute dict.  Remove key if
                # found so that if multiple keys maps to one attribute dict, the
                # next iteration of the outer loop finds the next key:
                for k0 in d0_all:
                    if d0_all[k0] == d0:
                        del d0_all[k0]
                        break
                for k1 in d1_all:
                    if d1_all[k1] == d1:
                        del d1_all[k1]
                        break
                d = deepdiff.DeepDiff(g0.get_edge_data(i0, j0, k0),
                                      g1.get_edge_data(i1, j1, k1))
                if d:
                    edge_diff[((i0, j0, k0), (i1, j1, k1))] = d
    else:
        raise ValueError('graph type not yet supported')
    return node_diff, edge_diff

def read_gexf(path):
    """
    Read graph in GEXF format from path.

    Reads a graph from a GEXF file using `networkx.read_gexf()`, but strips
    'id' and 'label' attributes from all nodes and edges.

    Parameters
    ----------
    path : file or string
        File or file name to write.
        File names ending in .gz or .bz2 will be compressed.    

    Returns
    -------
    graph : NetworkX graph
        If no parallel edges are found a Graph or DiGraph is returned.
        Otherwise a MultiGraph or MultiDiGraph is returned.
    """

    g = nx.read_gexf(path)
    for n in g.nodes():
        try:
            del g.node[n]['id']
        except:
            pass
        try:
            del g.node[n]['label']
        except:
            pass
    if isinstance(g, nx.MultiDiGraph):
        for n, m, k in g.edges(keys=True):
            try:
                del g.edge[n][m][k]['id']
            except:
                pass
    else:
        for n, m in g.edges():
            try:
                del g.edge[n][m]['id']
            except:
                pass

    return g



#!/usr/bin/env python

"""
Create generic LPU and simple pulse input signal.
"""

from itertools import product
import sys

import numpy as np
import h5py
import networkx as nx

def create_lpu_graph(lpu_name, N_sensory, N_local, N_proj):
    """
    Create a generic LPU graph.

    Creates a graph containing the neuron and synapse parameters for an LPU
    containing the specified number of local and projection neurons. The graph
    also contains the parameters for a set of sensory neurons that accept
    external input. All neurons are either spiking or graded potential neurons;
    the Leaky Integrate-and-Fire model is used for the former, while the
    Morris-Lecar model is used for the latter (i.e., the neuron's membrane
    potential is deemed to be its output rather than the time when it emits an
    action potential). Synapses use either the alpha function model or a
    conductance-based model.

    Parameters
    ----------
    lpu_name : str
        Name of LPU. Used in port identifiers.
    N_sensory : int
        Number of sensory neurons.
    N_local : int
        Number of local neurons.
    N_proj : int
        Number of project neurons.

    Returns
    -------
    g : networkx.MultiDiGraph
        Generated graph.
    """

    # Set numbers of neurons:
    neu_type = ('sensory', 'local', 'proj')
    neu_num = (N_sensory, N_local, N_proj)

    # Neuron ids are between 0 and the total number of neurons:
    G = nx.DiGraph()
    G.add_nodes_from(range(sum(neu_num)))

    idx = 0
    spk_out_id = 0
    gpot_out_id = 0
    for (t, n) in zip(neu_type, neu_num):
        for i in range(n):
            name = t+"_"+str(i)

            # All local neurons are graded potential only:
            if t != 'local' and np.random.rand() < 0.5:
                G.node[idx] = {
                    'model': 'LeakyIAF',
                    'name': name+'_s',
                    'extern': True if t == 'sensory' else False, # True if the neuron can receive external input
                    'public': True if t == 'proj' else False,    # True if the neuron can emit output
                    'spiking': True,
                    'V': np.random.uniform(-0.06,-0.025),
                    'Vr': -0.0675489770451,
                    'Vt': -0.0251355161007,
                    'R': 1.02445570216,
                    'C': 0.0669810502993}

                # Projection neurons are all assumed to be attached to output
                # ports (which are not represented as separate nodes):
                if t == 'proj':
                    G.node[idx]['selector'] = '/%s/out/spk/%s' % (lpu_name, str(spk_out_id))
                    G.node[idx]['circuit'] = 'proj'
                    spk_out_id += 1
                else:
                    G.node[idx]['circuit'] = 'local'
            else:
                G.node[idx] = {
                    'model': "MorrisLecar",
                    'name': name+'_g',
                    'extern': True if t == 'sensory' else False,
                    'public': True if t == 'proj' else False,
                    'spiking': False,
                    'V1': 0.03,
                    'V2': 0.015,
                    'V3': 0,
                    'V4': 0.03,
                    'phi': 0.025,
                    'offset': 0,
                    'initV': -0.05214,
                    'initn': 0.02,
                }

                # Projection neurons are all assumed to be attached to output
                # ports (which are not represented as separate nodes):
                if t == 'proj':
                    G.node[idx]['selector'] = '/%s/out/gpot/%s' % (lpu_name, str(gpot_out_id))
                    G.node[idx]['circuit'] = 'proj'
                    gpot_out_id += 1
                else:
                    G.node[idx]['circuit'] = 'local'
            idx += 1

    # An input port node is created for and attached to each non-projection
    # neuron with a synapse; this assumes that data propagates from one LPU to
    # another as follows:
    # LPU0[projection neuron] -> LPU0[output port] -> LPU1[input port] -> 
    # LPU1[synapse] -> LPU1[non-projection neuron]
    spk_in_id = 0
    gpot_in_id = 0
    for i, data in G.nodes_iter(True):
        if data['public'] == False:
            if data['spiking']:
                G.add_node(idx, {
                    'name': 'port_in_spk_%s' % spk_in_id,
                    'model': 'port_in_spk',
                    'selector': '/%s/in/spk/%s' % (lpu_name, idx),
                    'spiking': True,
                    'public': False,
                    'extern': False,
                    'circuit': G.node[i]['circuit']
                })
                spk_in_id += 1
                G.add_edge(idx, i, **{
                    'name': G.node[idx]['name']+'-'+G.node[i]['name'],
                    'model': 'AlphaSynapse',
                    'class': 0,
                    'conductance': True,
                    'ad': 0.19*1000,
                    'ar': 1.1*100,
                    'gmax': 0.003,
                    'reverse': 0.065,
                    'circuit': G.node[i]['circuit']})
            else:
                G.add_node(idx, {
                    'name': 'port_in_gpot_%s' % i,
                    'model': 'port_in_gpot',
                    'selector': '/%s/in/gpot/%s' % (lpu_name, idx),
                    'spiking': False,
                    'public': False,
                    'extern': False,
                    'circuit': G.node[i]['circuit']
                })
                gpot_in_id += 1
                G.add_edge(idx, i, **{
                    'name': G.node[idx]['name']+'-'+G.node[i]['name'],
                    'model': 'power_gpot_gpot',
                    'class': 3,
                    'conductance': True,
                    'slope': 0.8,
                    'reverse': -0.08,
                    'saturation': 0.03,
                    'power': 1.0,
                    'delay': 1.0,
                    'threshold': -0.05,
                    'circuit': G.node[i]['circuit']})

        idx += 1

    # Assume a probability of synapse existence for each group of synapses:
    # sensory -> local, sensory -> projection, local -> projection, 
    # projection -> local:            
    for r, (i, j) in zip((0.5, 0.1, 0.1, 0.3),
                         ((0, 1), (0, 2), (1, 2), (2, 1))):
        src_off = sum(neu_num[0:i])
        tar_off = sum(neu_num[0:j])
        for src, tar in product(range(src_off, src_off+neu_num[i]),
                                range(tar_off, tar_off+neu_num[j])):

            # Don't connect all neurons:
            if np.random.rand() > r: continue

            # Connections from the sensory neurons use the alpha function model;
            # all other connections use the power_gpot_gpot model:
            name = G.node[src]['name'] + '-' + G.node[tar]['name']
            if G.node[src]['spiking'] is True:
                G.add_edge(src, tar, **{
                    'model'       : 'AlphaSynapse',
                    'name'        : name,
                    'class'       : 0 if G.node[tar]['spiking'] is True else 1,
                    'ar'          : 1.1*1e2,
                    'ad'          : 1.9*1e3,
                    'reverse'     : 65*1e-3 if G.node[tar]['spiking'] else 0.01,
                    'gmax'        : 3*1e-3 if G.node[tar]['spiking'] else 3.1e-4,
                    'conductance' : True,
                    'circuit'     : G.node[src]['circuit']})
            else:
                G.add_edge(src, tar, **{
                    'model'       : 'power_gpot_gpot',
                    'name'        : name,
                    'class'       : 2 if G.node[tar]['spiking'] is True else 3,
                    'slope'       : 0.8,
                    'threshold'   : -0.05,
                    'power'       : 1.0,
                    'saturation'  : 0.03,
                    'delay'       : 1.0,
                    'reverse'     : -0.1,
                    'conductance' : True,
                    'circuit'     : G.node[src]['circuit']})

    return G

def create_lpu(file_name, lpu_name, N_sensory, N_local, N_proj):
    """
    Create a generic LPU graph.

    Creates a GEXF file containing the neuron and synapse parameters for an LPU
    containing the specified number of local and projection neurons. The GEXF
    file also contains the parameters for a set of sensory neurons that accept
    external input. All neurons are either spiking or graded potential neurons;
    the Leaky Integrate-and-Fire model is used for the former, while the
    Morris-Lecar model is used for the latter (i.e., the neuron's membrane
    potential is deemed to be its output rather than the time when it emits an
    action potential). Synapses use either the alpha function model or a
    conductance-based model.

    Parameters
    ----------
    file_name : str
        Output GEXF file name.
    lpu_name : str
        Name of LPU. Used in port identifiers.
    N_sensory : int
        Number of sensory neurons.
    N_local : int
        Number of local neurons.
    N_proj : int
        Number of project neurons.

    Returns
    -------
    g : networkx.MultiDiGraph
        Generated graph.
    """

    g = create_lpu_graph(lpu_name, N_sensory, N_local, N_proj)
    nx.write_gexf(g, file_name)

def create_input(file_name, N_sensory, dt=1e-4, dur=1.0, start=0.3, stop=0.6, I_max=0.6):
    """
    Create input stimulus for sensory neurons in artificial LPU.

    Creates an HDF5 file containing input signals for the specified number of
    neurons. The signals consist of a rectangular pulse of specified duration
    and magnitude.

    Parameters
    ----------
    file_name : str
        Name of output HDF5 file.
    N_sensory : int
        Number of sensory neurons.
    dt : float
        Time resolution of generated signal.
    dur : float
        Duration of generated signal.
    start : float
        Start time of signal pulse.
    stop : float
        Stop time of signal pulse.
    I_max : float
        Pulse magnitude.
    """

    Nt = int(dur/dt)
    t  = np.arange(0, dt*Nt, dt)

    I  = np.zeros((Nt, N_sensory), dtype=np.float64)
    I[np.logical_and(t>start, t<stop)] = I_max

    with h5py.File(file_name, 'w') as f:
        f.create_dataset('array', (Nt, N_sensory),
                         dtype=np.float64,
                         data=I)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('lpu_file_name', nargs='?', default='generic_lpu.gexf.gz',
                        help='LPU file name')
    parser.add_argument('in_file_name', nargs='?', default='generic_input.h5',
                        help='Input file name')
    parser.add_argument('-s', type=int,
                        help='Seed random number generator')
    parser.add_argument('-l', '--lpu', type=str, default='gen',
                        help='LPU name')

    args = parser.parse_args()

    if args.s is not None:
        np.random.seed(args.s)
    dt = 1e-4
    dur = 1.0
    start = 0.3
    stop = 0.6
    I_max = 0.6
    neu_num = [np.random.randint(31, 40) for i in xrange(3)]

    create_input(args.in_file_name, neu_num[0], dt, dur, start, stop, I_max)
    create_lpu(args.lpu_file_name, args.lpu, *neu_num)



#!/usr/bin/env python

"""
Query wrapper with support for operators.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license
import collections
import numbers
import pprint
import re
from datetime import datetime
import json
import time
import sys
from warnings import warn

from pyorient.ogm import Config, Graph
from pyorient.exceptions import PyOrientCommandException
from pyorient.utils import get_hash

from neuroarch.utils import is_rid, iterable, chunks, class_method_timer
from neuroarch.diff import diff_nodes, diff_edges
from neuroarch.apply_diff import apply_node_diff, apply_edge_diff

class UnsupportedQueryLanguage(Exception):
    """
    Unsupported query language.
    """
    pass

class UnsupportedView(Exception):
    """
    Unsupported view type.
    """
    pass

QueryString = collections.namedtuple('QueryString', ['str', 'lang'])

from neuroarch.conv import nx,pd

class QueryWrapper(object):
    """
    Gremlin/OrientDB SQL query wrapper.

    This class encapsulates Gremlin and OrientDB SQL queries that return nodes with
    functionality that permits the query to be composed with other queries using
    various operators. The query mechanism can also extract edges between the
    requested nodes.

    Parameters
    ----------
    graph : pyorient.ogm.Graph
        Graph object.
    query : QueryString, tuple, or list
        If a QueryString, contains a single node query. If a tuple/list, must contain
        an operator followed by two operands, which in turn may either be
        tuples/lists formatted as previously described or QueryString instances.
    init_nodes : set or dict
        Node records with which to initialize the query cache. If dict, the keys
        are the record IDs and the values the records.
    execute : bool
        Whether to execute the query immediately upon instantiation.
    executed : bool
        Whether the query has been executed.
    edges : bool
        If True, extract edges between requested nodes.
    debug : bool
        Run in debug mode. If true, will print execution time.
    """

    @class_method_timer
    def __init__(self, graph, query, init_nodes=set(), execute=True, executed = False,
                 edges=False, debug = False):
        assert isinstance(graph, Graph)
        self._graph = graph

        if isinstance(query, QueryString):
            self._query = query
        else:
            assert (isinstance(query, list) or isinstance(query, tuple)) and \
                len(query) == 3
            self._query = query
        if isinstance(init_nodes, dict):
            self._nodes = init_nodes
        else:
            self._nodes = self._records_to_dict(init_nodes)
        self._edges = {}
        self._executed = executed
        self._edge_executed = False
        self.debug = debug
        if execute:
            self.execute(edges)

    def __len__(self):
        if not self._executed:
            self.execute()
        return len(self._nodes)

    def __iter__(self):
        return self.node_objs.__iter__()

    @classmethod
    def empty_query(cls, graph, debug = False):
        return cls(graph, QueryString('select from #-1:0', 'sql'),
                   executed = True, execute = False, debug = debug)

    @classmethod
    def from_objs(cls, graph, objs, debug = False):
        """
        Construct a QueryWrapper given a list of OGM objects

        Parameters
        ----------
        graph : pyorient.ogm.Graph
            Graph object.
        obj : list
            List of orientdb objs

        Returns
        -------
        result : neuroarch.query.QueryWrapper
            QueryWrapper instance.
        """
        if len(objs) == 0:
            return cls.empty_query(graph, debug = debug)
            # return cls(graph,
            #            QueryString(str="""select from DataSource where name = "uiyth" """,
            #                        lang = 'sql'),
            #            init_nodes = set(), executed = True, debug = debug)

        rids = list(map(lambda x: x._id, objs)).__repr__().replace("'","")
        return cls(graph, QueryString(str='select from '+rids,
                                      lang='sql'),
                   debug = debug)

    @classmethod
    def from_tag(self, graph,tag, debug = False):
        """
        """
        obj = graph.QueryResults.query(tag=tag).all()
        if not obj:
            return {}
        QueryResultNode = QueryWrapper.from_objs(graph,obj)
        output = {}
        output['metadata'] = list(QueryResultNode.get_as('df')[0].to_dict(orient='index').values())[0]
        # for i in range(len(output['metadata']['pinned'])):
        #     output['metadata']['pinned'][i] = '#' + ":".join(map(str,output['metadata']['pinned'][i][1:]))
        output['qw'] = QueryResultNode.gen_traversal_out(['HasQueryResults'],min_depth=1)
        return output

    @classmethod
    def from_rids(cls, graph, *rid_list, **kwargs):
        """
        Construct a QueryWrapper given a list of OrientDB RIDs.

        Parameters
        ----------
        graph : pyorient.ogm.Graph
            Graph object.
        rid_list : list
            List of RIDs expressed as strings, e.g., '#12:0'.

        Returns
        -------
        result : neuroarch.query.QueryWrapper
            QueryWrapper instance.
        """

        if len(rid_list) == 0:
            return cls.empty_query(graph, debug = kwargs.get('debug', False))
            # return cls(graph,
            #            QueryString(str="""select from DataSource where name = "uiyth" """,
            #                        lang = 'sql'),
            #            init_nodes = set(), executed = True,
            #            debug = kwargs.get('debug', False))
        assert all([is_rid(rid) for rid in rid_list])
        return cls(graph, QueryString(str='select from [%s]' % ','.join(rid_list),
                                      lang='sql'),
                   debug = kwargs.get('debug', False))

    @classmethod
    def from_elements(cls, graph, *obj_list, **kwargs):
        """
        Construct a QueryWrapper given a list of pyorient OGM objects.

        Parameters
        ----------
        graph : pyorient.ogm.Graph
            Graph object.
        obj_list : list of neuroarch.models.Node or neuroarch.models.Relationship
            List of pyorient OGM objects.

        Returns
        -------
        result : neuroarch.query.QueryWrapper
            QueryWrapper instance.
        """
        if len(obj_list) == 0:
            return cls.empty_query(graph, debug = kwargs.get('debug', False))
            # return cls(graph,
            #            QueryString(str="""select from DataSource where name = "uiyth" """,
            #                        lang = 'sql'),
            #            init_nodes = set(), executed = True,
            #            debug = kwargs.get('debug', False))
        return cls(graph, QueryString(str='select from [%s]' % \
                                      ','.join([obj._id for obj in obj_list]),
                                      lang='sql'),
                   debug = kwargs.get('debug', False))

    @classmethod
    def _records_to_dict(cls, records):
        return {r._rid: r for r in records}

    @classmethod
    def _records_to_list(cls, records):
        return [r._rid for r in records]

    @class_method_timer
    def get_as(self, as_type='df', force_rid=False, edges = True, edge_class = '', deepcopy = False):
        """
        Return view of query results.

        Parameters
        ----------
        as_type : {'df', 'nx', 'obj'}
            If 'df', return results as a tuple of Pandas DataFrame instances.
            If 'nx', return a single NetworkX MultiDiGraph instance.
            If 'obj', return a tuple of pyorient OGM object lists.
        force_rid : bool
            If True, always use the OrientDB RID as the node identifier or index
            in the returned graph or DataFrame. Otherwise, use 'id' property.
            Ignored if `as_type` == 'obj'.

        Returns
        -------
        result : tuple or networkx.MultiDiGraph
            Query results in specified format.
        """

        if edges:
            self.execute_edges(edge_class = edge_class)

        if as_type=='df':
            return pd.as_pandas(self.nodes, self.edges, force_rid, deepcopy = deepcopy)
        if as_type=='nx':
            tmp = nx.as_nx(self.nodes, self.edges, force_rid, deepcopy = deepcopy)
            return tmp
        if as_type=='obj':
            return self._graph.elements_from_records(self.nodes),\
                self._graph.elements_from_records(self.edges)
        raise UnsupportedView(as_type)

    @class_method_timer
    def get_nodes(self, as_type = 'obj', force_rid = False, deepcopy = False):
        """
        Return view of nodes in query results.

        Parameters
        ----------
        as_type : {'df', 'obj'}
            If 'df', return nodes as a Pandas DataFrame instances.
            If 'obj', return a iist of pyorient OGM objects.
        force_rid : bool
            If True, always use the OrientDB RID as the node identifier or index
            in the returned graph or DataFrame. Otherwise, use 'id' property.
            Ignored if `as_type` == 'obj'.

        Returns
        -------
        result :
            Nodes in the query results in specified format.
        """
        if as_type=='obj':
            return self._graph.elements_from_records(self.nodes)
        elif as_type=='df':
            return pd.as_pandas(nodes = self.nodes, force_rid = force_rid,
                                deepcopy = deepcopy)[0]
        raise UnsupportedView(as_type)

    @property
    def executed(self):
        """
        Whether the query has been executed.
        """

        return self._executed

    @property
    def query_string(self):
        """
        The encapsulated query string(s).
        """

        return self._query

    @property
    def node_rids(self):
        if not self._executed:
            self.execute()
        return list(set(self._nodes.keys()))

    @property
    def nodes(self):
        """
        Set of node records retrieved by query.
        """

        if not self._executed:
            self.execute()
        return set(self._nodes.values())

    @property
    def nodes_as_list(self):
        """
        List of node records retrieved by query.
        """

        if not self._executed:
            self.execute()
        return list(self._nodes.values())

    @property
    def node_objs(self):
        """
        List of pyorient objects corresponding to node records.
        This property do not retrieve again data from the database.
        """
        return self._graph.elements_from_records(self.nodes)

    @property
    def nodes_as_objs(self):
        """
        List of pyorient objects corresponding to node records retrieved by query.
        This is different from QueryWrapper.node_objs in that this property retrieves
        the nodes from database, whereas QueryWrapper.node_objs uses the records stored in self.nodes.
        """
        if not self._executed:
            self.execute()
        return list(map(self._graph.get_element, self._nodes))

    @property
    def edges(self):
        """
        Set of edge records between nodes retrieved by query.
        """

        if not self._executed:
            self.execute()
        return set(self._edges.values())

    @property
    def edges_as_list(self):
        """
        List of edge records between nodes retrieved by query.
        """

        self.execute(edges = True)
        return list(self._edges.values())

    @property
    def edges_as_objs(self):
        """
        List of pyorient objects corresponding to edge records between nodes retrieved by query.
        """

        self.execute(edges = True)
        return list(map(self._graph.get_element, self._edges))

    def __nonzero__(self):
        return True if self._nodes else False

    def __repr__(self):
        return ('QueryWrapper\n------------\n'
                'Nodes: [%i]\n'
                'Edges: [%i]\n') % (len(self._nodes), len(self._edges))

    def _execute_query(self, q):
        """
        Return results of a single OrientDB query string.
        """

        assert isinstance(q, QueryString)
        if q.lang == 'gremlin':
            result = self._graph.client.gremlin(q.str)
        elif q.lang == 'sql':
            result = self._graph.client.command(q.str)
        else:
            raise UnsupportedQueryLanguage(q.lang)
        return result

    def _edge_query_from_node_query(self, node_query, edge_class=''):
        """
        Return query string that can retrieve edges between nodes selected by specified query string.

        Parameters
        ----------
        node_query : QueryString
            Query that returns node.
        edge_class : str or iterable of str
            Name of edge class with which to restrict the constructed edge
            query. Multiple classes may be specified. If no classes are
            specified, nodes of any class are retrieved.

        Returns
        -------
        edge_query : QueryString
            Query that retrieves edges between nodes selected by `node_query`.
        """

        assert isinstance(node_query, QueryString)
        if not edge_class:
            edge_class_str = ''
        elif isinstance(edge_class, str):
            edge_class_str = '"%s"' % str(edge_class)
        elif iterable(edge_class):
            edge_class_str = ','.join([str(c) for c in edge_class])
        else:
            raise ValueError('invalid edge class')
        if node_query.lang == 'gremlin':
            s = 'x='+'.'.join((node_query, 'outE(%s).toList();' % edge_class_str))+\
                'y='+'.'.join((node_query, 'inE(%s).toList();' % edge_class_str))+\
                'x.intersect(y)'
            return QueryString(str=s, lang='gremlin')
        elif node_query.lang == 'sql':
            s = ("""select expand($c) let $a = (select expand(oute(%s)) from (%s)),"""
                 """$b = (select expand(ine(%s)) from (%s)),"""
                 """$c = intersect($a,$b)""") % (str(edge_class_str),
                                             str(node_query), str(edge_class_str),
                                             str(node_query))
            return QueryString(str=s, lang='sql')
        else:
            raise UnsupportedQueryLanguage(node_query.lang)

    def _edge_query_from_node_rids(self, node_rids, edge_class=''):
        """
        Return query string that can retrieve edges between specified node IDs.

        Parameters
        ----------
        node_rids : list of str
            List of node IDs.
        edge_class : str or iterable of str
            Name of edge class with which to restrict the constructed edge
            query. Multiple classes may be specified. If no classes are specified,
            nodes of any class are retrieved.

        Returns
        -------
        edge_query : QueryString
            Query that retrieves edges between nodes in `node_rids`.
        """

        if len(node_rids) == 0:
            return QueryString(str="""select from DataSource where name = "uiyth" """,
                                   lang = 'sql')
        if not edge_class:
            edge_class_str = ''
        elif isinstance(edge_class, str):
            edge_class_str = '"%s"' % edge_class
        elif iterable(edge_class):
            edge_class_str = ','.join([str(c) for c in edge_class])
        else:
            raise ValueError('invalid edge class')
        assert iterable(node_rids)
        rid_str = ','.join(['%s' % rid for rid in node_rids])
        s = ("""select expand($c) let $a = (select expand(oute(%s)) from [%s]),"""
             """$b = (select expand(ine(%s)) from [%s]),"""
             """$c = intersect($a,$b)""") % (str(edge_class_str), str(rid_str),
                                         str(edge_class_str), str(rid_str))
        return QueryString(str=s, lang='sql')

    def _connect_to_query_result_node(self, retries=1):
        """
        Create a new `query_result` node and connect it to all of the cached
        nodes of the query via `query_owns` edges.
        """
        # FIXME
        if not self._result:
            return
        batch = self._graph.batch()
        batch['query_result_node'] = batch.query_results.create(query=self._query)
        for i, rid in enumerate(self._nodes):
            batch[str(i)] = batch.query_owns.create(batch['query_result_node'],
                                                    self._graph.get_element(rid))
        batch.commit(retries)



    @staticmethod
    def _dict_intersection(a, b):
        return {k:a[k] for k in set(a).intersection(b)}

    @staticmethod
    def _dict_union(a, b):
        c = a.copy()
        c.update(b)
        return c

    @staticmethod
    def _dict_difference(a, b):
        return {k:a[k] for k in set(a).difference(b)}

    @staticmethod
    def _dict_symmetric_difference(a, b):
        return {k:(a[k] if k in a else b[k]) for k in \
                set(a).symmetric_difference(b)}

    @classmethod
    def multi_traverse_owned_by_toplevel(cls, queryObjList):
        if isinstance(queryObjList, cls):
            queryObjList=list(queryObjList)
        assert isinstance(queryObjList, list)
        for idx, obj in enumerate(queryObjList):
            assert isinstance(obj, cls)
            if idx==0:
                 queryObj=obj
            else:
                queryObj=queryObj+obj
        return queryObj.traverse_owned_by_get_toplevel()

    @class_method_timer
    def execute(self, edges=False, force=False, connect=False):
        """
        Execute the query.

        Parameters
        ----------
        edges : bool
            If True, retrieve edges between nodes as well as nodes.
        force : bool
            Execute the query even if node/edge results have already been
            cached.
        connect : bool
            Create new `query_results` node and connect it to the nodes
            returned by the query.
        """

        def e(q):
            if isinstance(q, QueryString):
                result = self._records_to_dict(self._execute_query(q))
            else:
                if len(q) == 3:
                    r1 = e(q[1])
                    r2 = e(q[2])

                    if q[0] in ['*', '&']:
                        result = self._dict_intersection(r1, r2)
                    elif q[0] in ['+', '|']:
                        result = self._dict_union(r1, r2)
                    elif q[0] == '-':
                        result = self._dict_difference(r1, r2)
                    elif q[0] == '^':
                        result = self._dict_symmetric_difference(r1, r2)
                    else:
                        raise ValueError('unrecognized operator')
                else:
                    raise ValueError('problematic query entry')
            return result

        # Don't execute query if results have already been cached:
        if self._executed and not force:
            if self._edge_executed:
                return
            else:
                if edges:
                    self._edges = \
                        self._records_to_dict(self._execute_query(self._edge_query_from_node_rids(self._nodes.keys())))
                    self._edge_executed = True
                else:
                    return
        else:
            self._nodes = e(self._query)

            if edges:
                if not self._edge_executed:
                    self._edges = \
                        self._records_to_dict(self._execute_query(self._edge_query_from_node_rids(self._nodes.keys())))
                    self._edge_executed = True
            if connect:
                self._connect_to_query_result_node(1)
            self._executed = True

    @class_method_timer
    def execute_edges(self, edge_class = ''):
        if self._executed:
            if self._edge_executed:
                return
            else:
                self._edges = \
                    self._records_to_dict(self._execute_query(
                        self._edge_query_from_node_rids(self._nodes.keys(), edge_class = edge_class)))
                self._edge_executed = True
        else:
            self.execute(edges = True)

    def clear(self):
        """
        Clear cached query results.
        """

        self._nodes = dict()
        self._edges = dict()
        self._executed = False

        # XXX what should be done with the query_result node in the db?

    @class_method_timer
    def owns(self, levels=1, **kwargs):
        """
        Retrieve all nodes at a specified number of levels below the current query.

        Parameters
        ----------
        levels : int
            Number of ownership edges between the current query and the nodes to
            be retrieved.
        cls : str or iterable
            Node class or classes to retrieve.

        Returns
        -------
        result : QueryWrapper
            Query containing requested nodes.
        """

        return self._own('out', levels, kwargs)

    @class_method_timer
    def owned_by(self, levels=1, **kwargs):
        """
        Retrieve all nodes at a specified number of levels above the current query.

        Parameters
        ----------
        levels : int
            Number of ownership edges between the current query and the nodes to
            be retrieved.
        cls : str or iterable
            Node class or classes to retrieve.

        Returns
        -------
        result : QueryWrapper
            Query containing requested nodes.
        """

        return self._own('in', levels, kwargs)

    def _own(self, direction, levels, kwargs):
        assert isinstance(levels, numbers.Integral) and levels >= 1

        rid_list = self.node_rids
        if len(rid_list) == 0:
            return self.empty_query(self._graph, debug = self.debug)
            # return self.__class__(self._graph, QueryString("""select from DataSource where name = "uiyth" ""","sql"),
            #                       debug = self.debug, edges = self.edges)
        classes, attrs, depth, columns = _kwargs(kwargs)

        relationships = ["""%s('owns')""" % direction]*levels
        query = """select %s from (select expand(%s) from [%s]) %s""" % \
                    (columns, '.'.join(relationships), ", ".join(rid_list), classes)
        '''
        disp_query = "select %s from (select expand(%s) from (%s)) %s" % \
                    (columns, '.'.join(relationships), self._disp_query, classes)
        '''
        disp_query = ''
        return self.__class__(self._graph, QueryString(query,"sql"), debug = self.debug)

    @class_method_timer
    def find_matching_ports_from_selector(self, other):
        try:
            other_df = other.get_as('df')[0][['selector']].to_dict(orient='index').values()
            sels = [list(a.values())[0] for a in other_df if isinstance(list(a.values())[0],str)]
        except KeyError:
            other_df = []
            sels = []
        return self.traverse_owns(selector=sels)

    @class_method_timer
    def traverse_owns(self, **kwargs):
        """
        Retrieve all traversed nodes down to a specified number of levels below the current query.

        Parameters
        ----------
        max_levels : int
            Number of ownership edges between the current query and the nodes to
            be retrieved.
        cls : str or iterable
            Node class or classes to retrieve.

        Returns
        -------
        result : QueryWrapper
            Query containing requested nodes.
        """

        return self._traverse('out', kwargs)

    @class_method_timer
    def traverse_owned_by(self, **kwargs):
        """
        Retrieve all traversed nodes up to a specified number of levels above the current query.

        Parameters
        ----------
        max_levels : int
            Number of ownership edges between the current query and the nodes to
            be retrieved.
        cls : str or iterable
            Node class or classes to retrieve.

        Returns
        -------
        result : QueryWrapper
            Query containing requested nodes.
        """

        return self._traverse('in', kwargs)

    def _traverse(self, direction, kwargs):
        if not 'max_levels' in kwargs:
            kwargs['max_levels']=10

        rid_list = self.node_rids
        if len(rid_list) == 0:
            return self.empty_query(self._graph, debug = self.debug)
            # return self.__class__(self._graph, QueryString("""select from DataSource where name = "uiyth" ""","sql"),
            #                       debug = self.debug, edges = self.edges)

        classes, attrs, depth, columns = _kwargs(kwargs)

        attrs_query = ""
        if attrs and classes:
            attrs_query = " and (" + " and ".join(attrs) + ") "
        elif (not classes) and attrs:
            attrs_query = " where (" + " and ".join(attrs) + ") "


        query = """select %s from (traverse %s('owns') from [%s] %s) %s %s""" %\
                (columns, direction, ", ".join(rid_list), depth, classes, attrs_query)
        '''
        disp_query = "select %s from (traverse %s('owns') from (%s) %s) %s %s" %\
                (columns, direction, self._disp_query, depth, classes, attrs_query)
        '''
        disp_query = ''
        return self.__class__(self._graph, QueryString(query, "sql"),
                              debug = self.debug)

    '''
    def traverse_owned_by_get_toplevel(self):
        """
        Find associated LPU or Pattern that of the nodes encompassed by a query object.

        Returns
        -------
        result : dict
            QueryWrappers broken out by LPU/Pattern containing nodes of the origin query
        """
        toplevel = dict()
        rid_list = self.node_rids

        for c in ('LPU', 'Pattern'):
            obj=self.traverse_owned_by(cls=c, max_levels=10)
            if c not in toplevel:
                toplevel[c] = dict()
            for n in obj.nodes_as_objs:
                query = "select @RID as rid from (traverse out('owns') from [%s] while $depth <= 10)" % (n._id)
                trav_owns = n._graph.client.command(query)
                if trav_owns and isinstance(trav_owns[0].oRecordData['rid'],tuple):
                    obj_rid_list = set(rid_list) & set(['#' + str(record.oRecordData['rid'][1]) + ':' + str(record.oRecordData['rid'][2])
                                                        for record in trav_owns])
                else:
                    obj_rid_list = set(rid_list) & set([record.oRecordData['rid'].get_hash() for record in trav_owns])
                if n.name not in toplevel['LPU']:
                    toplevel[c][n.name]=set()
                toplevel[c][n.name] |= obj_rid_list

        for key, v in toplevel.items():
            for name, rids in v.items():
                toplevel[key][name] = self.from_rids(self._graph, *rids)
        return toplevel
    '''

    @class_method_timer
    def traverse_owned_by_get_toplevel(self):
        """
        Find associated LPU or Pattern that of the nodes encompassed by a query object.

        Returns
        -------
        result : dict
            QueryWrappers broken out by LPU/Pattern containing nodes of the origin query
        """
        toplevel = dict()
        rid_list = self.node_rids

        for c in ('LPU', 'Pattern'):
            obj=self.traverse_owned_by(cls=c, max_levels=10)
            if c not in toplevel:
                toplevel[c] = dict()
            for n in obj.node_objs:
                query = "select expand($c) let $a=(select from (traverse out('owns') from %s while $depth <= 10))" % (n._id)
                query += ", $b=(select from [%s]), $c=intersect($a,$b)" % (", ".join(rid_list))
                toplevel[c][n.name] = self.__class__(self._graph, QueryString(query,'sql'), debug = self.debug)
        return toplevel

    @class_method_timer
    def get_data_rids(self, as_type='df', **kwargs):
        if len(self._nodes) == 0:
            return []

        rid_list = self.node_rids
        classes, attrs, depth, columns = _kwargs(kwargs)
        attrs_query = ""
        if attrs and classes:
            attrs_query = " and (" + " and ".join(attrs) + ") "
        elif (not classes) and attrs:
            attrs_query = " where (" + " and ".join(attrs) + ") "
        elif classes and not attrs:
            attrs_query = ""

        columns = '@rid as rid'
        query = "select %s from (select expand(out('HasData')) from [%s]) %s %s" % \
                (columns, ", ".join(rid_list), classes, attrs_query)

        res = self._graph.client.command(query)
        # if res and isinstance(res[0].oRecordData['rid'],tuple):
        #     res = ['#' + str(record.oRecordData['rid'][1]) + ':' + str(record.oRecordData['rid'][2])
        #            for record in res]
        # else:
        #     res = [record.oRecordData['rid'].get_hash() for record in res]
        res = [record._rid for record in res]
        return res

    @class_method_timer
    def get_data(self, as_type='df', edges = True, deepcopy = True, **kwargs):
        rid_list = self.node_rids
        if len(rid_list) == 0:
            if as_type == 'df':
                return pd.pd.DataFrame()
            elif as_type == 'obj':
                return ([],[])
            elif as_type == 'nx':
                return nx.nx.MultiDiGraph()

        classes, attrs, depth, columns = _kwargs(kwargs)
        attrs_query = ""
        if attrs and classes:
            attrs_query = " and (" + " and ".join(attrs) + ") "
        elif (not classes) and attrs:
            attrs_query = " where (" + " and ".join(attrs) + ") "


        query = "select %s from (select expand(out('HasData')) from [%s]) %s %s" % \
                    (columns, ", ".join(rid_list), classes, attrs_query)

        return QueryWrapper(self._graph, QueryString(query, 'sql'), debug = self.debug).get_as(
                                    as_type, edges = edges, deepcopy = deepcopy)


    @class_method_timer
    def get_data_qw(self, **kwargs):
        rid_list = self.node_rids
        if len(rid_list) == 0:
            return self.empty_query(self._graph, debug = self.debug)

        classes, attrs, depth, columns = _kwargs(kwargs)
        attrs_query = ""
        if attrs and classes:
            attrs_query = " and (" + " and ".join(attrs) + ") "
        elif (not classes) and attrs:
            attrs_query = " where (" + " and ".join(attrs) + ") "


        query = "select %s from (select expand(out('HasData')) from [%s]) %s %s" % \
                    (columns, ", ".join(rid_list), classes, attrs_query)

        return QueryWrapper(self._graph, QueryString(query, 'sql'), debug = self.debug)

    def query(self, **kwargs):
        self.has(**kwargs)

    @class_method_timer
    def has(self, **kwargs):
        if not kwargs:
            return self

        q={}
        rid_list = self.node_rids
        if not len(rid_list):
            return self
        classes, attrs, depth, columns = _kwargs(kwargs)

        q_str = "{var} = (select expand(rid) from (select distinct(traversedvertex(0)) as rid \
                from (traverse out('Models'), out('HasData') from (select from [{rids}] {classes}) \
                while $depth <= 2) {filters}))"
        dq_str = "{var} = (select expand(rid) from (select distinct(traversedvertex(0)) as rid \
                from (traverse out('Models'), out('HasData') from ({disp_query} {classes}) \
                while $depth <= 2) {filters}))"
        dq={}
        if attrs:
            for i, a in enumerate(attrs):
                filters = "where " + a
                var = '$q'+str(i)
                q[var] = q_str.format(var = var, rids = ", ".join(rid_list), classes = classes, filters = filters)
                #dq[var] = dq_str.format(var = var, disp_query = self._disp_query, classes = classes, filters = filters)
        else:
            var = '$q'
            q['$q'] = q_str.format(var = var, rids = ", ".join(rid_list), classes = classes, filters = "")
            #dq['$q'] = dq_str.format(var = var, disp_query = self._disp_query, classes = classes, filters = "")

        query = """select %s from (select expand($a) let %s, $a = intersect(%s))""" % \
                    (columns, ", ".join(list(q.values())), ", ".join(list(q.keys())) )
        '''
        disp_query = "select %s from (select expand($a) let %s, $a = intersect(%s))" % \
                    (columns, ", ".join(dq.values()), ", ".join(dq.keys()) )
        '''
        disp_query = ''
        #print disp_query
        return self.__class__(self._graph, QueryString(query,"sql"), debug = self.debug)


    def _check_tags(self, tag):
        query = "select tag from QueryResult where tag = '%s'" % tag
        #print query
        results = self._graph.client.command(query)
        return self._records_to_list(results)

    @class_method_timer
    def tag_query_result_node(self, tag, permanent_flag, **kwargs):
        """
        Create a new `QueryResult` node and connect it to all of the cached
        nodes of the query via `HasQueryResults` edges.
        """
        # check if tag already exists
        if self._check_tags(tag): return -1
        qr = self._graph.QueryResults.create(tag=tag)
        kwargs['tag'] = tag
        self._graph.client.command('update %s content %s where @rid = %s' % \
                                     (qr.element_type, json.dumps(kwargs), qr._id))

        cmd = ['begin']

        # create tag node
        #set_cmd = ["%s = %s" % (k, v.__repr__()) for k, v in kwargs.items()]
        #if set_cmd:
        #    a = [""]
        #    a.extend(set_cmd)
        #    set_cmd = a
        #let_cmd = "let v = CREATE VERTEX QueryResult SET tag = '%s', permanent = %s, created_timestamp = sysdate() %s" % \
        #            (tag, permanent_flag, ", ".join(set_cmd))
        #cmd.append(let_cmd)

        # create edges from tag node to query nodes
        for i, node in enumerate(self._nodes):
            edge_cmd = "let e%s = CREATE EDGE HasQueryResults FROM %s TO %s" % (i, qr._id, node)
            cmd.append(edge_cmd)

        cmd.append("commit retry 10;\nreturn $v;")
        results = self._graph.client.batch(";\n".join(cmd))
        return 1
        #return self.from_rids(self._graph, *self._records_to_list(results))

    def tag_clean_up(self, older_than):
        cmd = "begin; DELETE VERTEX QueryResults WHERE permanent=False and created_timestamp <= DATE(%s); commit retry 10;" % \
                (older_than)
        results = self._graph.client.batch(cmd)

    def _get_node_num(self, node_name, node_class):
        col_name = re.sub('[#0-9]', '', node_name)
        query = "select max(name.replace('%s', '').asInteger()) from %s where name like '%s%%'" % \
                (col_name, node_class, col_name)
        res = self._graph.client.command(query)
        return res[0].oRecordData['max']

    def _get_subgraphs(self, edge_types=None, max_levels=10, max_node_cls=None, cls=None):
        return self._get_graphs('out', edge_types, max_levels, max_node_cls, cls)

    def _get_supgraphs(self, edge_types=None, max_levels=10, max_node_cls=None, cls=None):
        return self._get_graphs('in', edge_types, max_levels, max_node_cls, cls)

    def _get_graphs(self, direction, edge_types=None, max_levels=10, max_node_cls=None, cls=None):
        rid_list = self.node_rids
        class_list = list(self._graph.registry.keys())

        assert isinstance(max_levels, numbers.Integral) and max_levels >= 0

        if cls:
            assert cls in class_list
            cls = _list_repr(cls)
            where_classes = "where @class in %s" % cls
        else:
            where_classes = ""

        if edge_types:
            edge_types = _list_repr(edge_types)
            assert all(e in class_list for e in edge_types)
            relationships = ["%s('%s')" % (direction, e) for e in edge_types]
        else:
            relationships = ["in()"]

        if max_node_cls:
            assert max_node_cls in class_list
            while_classes = "and @class <> '%s'" % max_node_cls
        else:
            while_classes = ""

        query = "select from (traverse %s from [%s] while $depth<=%s %s) %s" % \
                (", ".join(relationships), ", ".join(rid_list), max_levels, while_classes, where_classes)
        disp_query = ''
        '''
        disp_query = "select from (traverse %s from (%s) while $depth<=%s %s) %s" % \
                (", ".join(relationships), self._disp_query, max_levels, while_classes, where_classes)
        '''

        return self.__class__(self._graph, QueryString(query, "sql"), debug = self.debug)

    def _get_in_edges(self, rid, edge_types):
        in_relationships = ["in('%s') as %s" % (e, e) for e in edge_types]
        if isinstance(rid, str):
            rids = [rid]
        query = "select @RID as rid, %s from %s " % (", ".join(in_relationships), ", ".join(rids))


        records = self._graph.client.command(query)
        in_edges = dict()
        for r in records:
            # rid_hash = r.oRecordData['rid'].get_hash()
            rid_hash = r._rid
            if rid_hash not in in_edges:
                in_edges[rid_hash] = dict()
            for edge in edge_types:
                in_edges[rid_hash][edge] = [e.get_hash() for e in r.oRecordData[edge]]

        return in_edges

    def _in_edges_command(self, to_from_records, edge_types):
        edge_in = "let f{num} = CREATE EDGE {edge_type} from {from_nodes} to {to_node}"
        in_cmd = []

        for i, (to_rid, from_rid) in enumerate(to_from_records):
            in_edges = self._get_in_edges(to_rid, edge_types)
            for n, edge_type in enumerate(edge_types):
                from_nodes = in_edges[to_rid][edge_type]
                if from_nodes:
                    in_cmd.append(edge_in.format(to_node = to_rid, from_nodes = from_rid,
                                                  num = len(to_from_records) * i + n , edge_type = edge_type))
        return in_cmd

    def _rename_nodes(self, func, rid_list, cmd):
        max_node_num = dict()

        # rename new top-level nodes
        for i, rid in enumerate(rid_list):
            node = self._nodes[rid]
            old_name = node.oRecordData['name']
            node_name = re.sub('[#0-9]', '', old_name)

            if node_name not in max_node_num:
                max_node_num[node_name] = self._get_node_num(old_name, node._class)
            assert func(max_node_num[node_name]) > max_node_num[node_name], 'Node name must not already exist'

            max_node_num[node_name] = func(max_node_num[node_name])
            new_name = node_name + str(max_node_num[node_name])
            cmd = cmd.replace(old_name, new_name)
        return cmd

    def _copy_graph_command(self, edge_types=None):
        node_cmd, node_map = self._copy_node_command(commit_stmt=False)
        edge_cmd = self._copy_edge_command(node_map, edge_types=None, commit_stmt=False)

        cmd = 'begin;\n' + "".join(node_cmd) + "".join(edge_cmd) + ";\ncommit retry 100;\nreturn [%s];" % (", ".join(list(node_map.values())))
        return cmd, node_map


    def _copy_node_command(self, commit_stmt=False, N=20):
        node_map_full = dict()
        cmd_list = []
        num = 0

        # batch commands for nodes
        for chunk in chunks(self._nodes.items(), N):
            node_map = dict()
            cmd = []

            for rid, node in chunk:
                set_cmd = ["%s = %s" % (k, v.__repr__()) for k, v in node.oRecordData.items() \
                           if isinstance(v, (str, numbers.Number))]
                let_cmd = "let v%s = CREATE VERTEX %s SET " % (num, node._class)
                node_map[rid] = '$v%s' % num
                cmd.append(let_cmd + ", ".join(set_cmd))
                num += 1

            cmd = ";\n".join(cmd) + ";\n"

            if commit_stmt:
                cmd += "commit retry 100;\nreturn [%s];" % (", ".join(list(node_map.values())))
            cmd_list.append(cmd)
            node_map_full.update(node_map)

        return cmd_list, node_map_full

    def _copy_edge_command(self, node_map, edge_types=None, commit_stmt=False, N=20):
        cmd_list = []
        num = 0

        # batch commands for edges, given node variables
        for chunk in chunks(self.edges, N):
            cmd = []
            for edge in chunk:
                edge_cmd = "let e%s = CREATE EDGE %s from %s to %s" % \
                                    (num, edge._class, node_map[edge._out.__str__()], node_map[edge._in.__str__()])
                if edge_types:
                    if edge._class in edge_types:
                        cmd.append(edge_cmd)
                else:
                    cmd.append(edge_cmd)
                num += 1

            cmd = ";\n".join(cmd) + ";"

            if commit_stmt:
                cmd += "\ncommit retry 100;"
            cmd_list.append(cmd)

        return cmd_list

    def copy_models(self, func, in_flag, copies=1000):
        edge_types = ['Owns', 'SendsTo', 'HasData']
        rid_list = self.node_rids
        data = self._get_subgraphs(edge_types)
        results = list()

        for c in range(copies):
            cmd, node_map = data._copy_graph_command(edge_types)
            cmd  = self._rename_nodes(func, rid_list, cmd)

            # create in edges to subgraph if in_flag=True
            if in_flag:
                in_cmd = self._in_edges_command(zip(rid_list, [node_map[r] for r in rid_list]), edge_types)
                cmd = cmd.replace(";\ncommit", ";\n" + ";\n".join(in_cmd) + ";\ncommit")
            results += self._graph.client.batch(cmd)
            #print cmd + '\nNew records committed to database'

        return self.from_rids(self._graph, *self._records_to_list(results))

    @staticmethod
    def _remove_edges_by_node(df, del_nodes):
        del_filter = (~df['in'].isin(del_nodes)) & (~df['out'].isin(del_nodes))
        return df[del_filter]


    def _diff(self, new_df_nodes, new_df_edges, full_replace, node_map={}):
        old_df_nodes, old_df_edges = self.get_as()

        if node_map:
            old_df_nodes.index = [node_map[r] for r in old_df_nodes.index]
            new_df_nodes.index = [node_map[r] for r in new_df_nodes.index]
            for direction in ('in', 'out'):
                old_df_edges[direction] = old_df_edges[direction].map(lambda x: node_map[x])
                new_df_edges[direction] = new_df_edges[direction].map(lambda x: node_map[x])

        d_n = diff_nodes(old_df_nodes, new_df_nodes, full_replace)

        # filter out edges related to nodes that will be deleted
        del_nodes = list(d_n['del'].keys())
        old_df_edges_filter = self._remove_edges_by_node(old_df_edges, del_nodes)
        new_df_edges_filter = self._remove_edges_by_node(new_df_edges, del_nodes)

        d_e = diff_edges(old_df_edges_filter, new_df_edges_filter, full_replace)
        rid_list = apply_node_diff(self._graph.client, d_n)
        edge_rid_list = apply_edge_diff(self._graph.client, d_e)
        return rid_list

    def diff_save(self, new_df_nodes, new_df_edges, full_replace=False):
        rid_list = self._diff(new_df_nodes, new_df_edges, full_replace)
        return self.from_rids(self._graph, *rid_list)

    def diff_save_as(self, new_df_nodes, new_df_edges, **kwargs):
        class_list = list(self._graph.registry.keys())

        if 'max_levels_in' in kwargs:
            assert isinstance(kwargs['max_levels_in'], numbers.Integral) and kwargs['max_levels_in'] >= 0
        else:
            kwargs['max_levels_in'] = 0

        if 'max_levels_out' in kwargs:
            assert isinstance(kwargs['max_levels_out'], numbers.Integral) and kwargs['max_levels_out'] >= 0
        else:
            kwargs['max_levels_out'] = 0

        if 'max_node_cls' in kwargs:
            assert (kwargs['max_node_cls'] in class_list), "Assign new nodes to an existing class: \n%s" % \
                                                                ('\n'.join(class_list))
        else:
            kwargs['max_node_cls'] = None

        # copy graph
        q1 = self._get_supgraphs(max_levels=kwargs['max_levels_in'], max_node_cls=kwargs['max_node_cls'])
        q2 = self._get_subgraphs(max_levels=kwargs['max_levels_out'])
        q = q1 + q2

        # commit copy of nodes to database
        result_rids = []
        node_cmd_list, node_map = self._copy_node_command(commit_stmt=True)
        for node_cmd in node_cmd_list:
            results = self._graph.client.batch("begin;\n" + node_cmd)
            result_rids += [r._rid for r in results]

        # map old nodes to newly created nodes
        node_to_node_map = {k: v for k, v in zip(node_map.keys(), result_rids)}

        # commit copy of edges to database
        edge_cmd_list = self._copy_edge_command(node_to_node_map, commit_stmt=True)
        for edge_cmd in edge_cmd_list:
            self._graph.client.batch("begin;\n" + edge_cmd)

        # take diff, map to new nodes in db
        rid_list = self._diff(new_df_nodes, new_df_edges, full_replace = True, node_map = node_to_node_map)
        res = result_rids+rid_list
        return self.from_rids(self._graph, *res)

    @class_method_timer
    def post_synaptic_neurons(self, N=None, rel='>', include_inferred=True, include_fragments = False, high_prob = False):
        # N represents number of synapses
        #    if N is none, will return all post synaptic neurons
        #    else, it will only return postsynaptic neurons where the number of synapses
        #    satisfy <rel> N. See below for rel
        # rel can be '>'(default),'<','='
        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        if N:
            return self.gen_traversal_out(['SendsTo', synapse_classes, {'NHP' if high_prob else 'N':(rel,N)}],['SendsTo',neuron_classes], min_depth=2)
        else:
            if high_prob:
                return self.gen_traversal_out(['SendsTo', synapse_classes, {'NHP':(rel,0)}],['SendsTo', neuron_classes, 'instanceof'], min_depth=2)
            else:
                return self.gen_traversal_out(['SendsTo', synapse_classes],['SendsTo',neuron_classes, 'instanceof'], min_depth=2)

    @class_method_timer
    def _post_synaptic_neurons(self, N=None, rel='>', include_fragments = False, high_prob = False):
        # slower than using post_synaptic_neurons
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = {n._id: n for n in self.node_objs}
        return QueryWrapper(self._graph, QueryString(
            """select expand(target) from (MATCH {{class: Neuron, where: (@rid in [{rid}])}} -SendsTo-> {{class: Synapse{where}}} -SendsTo-> {{class: {neuron_class}, as: target}} return target)""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = ', where: ({} {} {})'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else ', where: (NHP {} {})'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ), 'sql'), debug = self.debug)
        

    @class_method_timer
    def pre_synaptic_neurons(self, N=None, rel='>', include_inferred=True, include_fragments = False, high_prob = False):
        # N represents number of synapses
        #    if N is none, will return all post synaptic neurons
        #    else, it will only return postsynaptic neurons where the number of synapses
        #    satisfy <rel> N. See below for rel
        # rel can be '>'(default),'<','='
        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        if N:
            return self.gen_traversal_in(['SendsTo', synapse_classes,{'NHP' if high_prob else 'N':(rel,N)}],['SendsTo',neuron_classes], min_depth=2)
        else:
            if high_prob:
                return self.gen_traversal_in(['SendsTo', synapse_classes, {'NHP':(rel,0)}],['SendsTo',neuron_classes, 'instanceof'], min_depth=2)
            else:
                return self.gen_traversal_in(['SendsTo', synapse_classes],['SendsTo',neuron_classes, 'instanceof'], min_depth=2)

    @class_method_timer
    def _pre_synaptic_neurons(self, N=None, rel='>', include_fragments = False, high_prob = False):
        # slower than using pre_synaptic_neurons
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = {n._id: n for n in self.node_objs}
        return QueryWrapper(self._graph, QueryString(
            """select expand(source) from (MATCH {{class: Neuron, where: (@rid in [{rid}])}} <-SendsTo- {{class: Synapse{where}}} -SendsTo-> {{class: {neuron_class}, as: source}} return source)""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = ', where: ({} {} {})'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else ', where: (NHP {} {})'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ), 'sql'), debug = self.debug)
    
    @class_method_timer
    def pre_synaptic_neurons_with_synapse_count(self, N=None, rel='>', include_inferred=True, include_fragments = False, high_prob = False):
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = {n._id: n for n in self.node_objs}
        links = self._graph.client.command(
            """MATCH {{class: {neuron_class}, where: (@rid in [{rid}]), as: post}} <-SendsTo- {{class: Synapse, {where} as: syn}} <-SendsTo- {{class: {neuron_class}, as: source}} return post, syn.{NHP} as N, source""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ))
        if include_inferred:
            links.extend(self._graph.client.command(
            """MATCH {{class: {neuron_class}, where: (@rid in [{rid}]), as: post}} <-SendsTo- {{class: InferredSynapse, {where} as: syn}} <-SendsTo- {{class: {neuron_class}, as: source}} return post, syn.{NHP} as N, source""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            )))
        out = {k: {} for k in rids}
        pres = set()
        for link in links:
            post = get_hash(link.oRecordData['post'])
            Nsyn = link.oRecordData['N']
            pre = get_hash(link.oRecordData['source'])
            out[post][pre] = Nsyn
            pres.add(pre)
        pre_neurons = {n._id: n for n in self.from_rids(self._graph, *list(pres)).node_objs}

        return [(rids[post], pre_neurons[pre], N) for post, v in out.items() for pre, N in v.items()]
    
    @class_method_timer
    def _pre_synaptic_neurons_with_synapse_count(self, N=None, rel='>', include_inferred=True, high_prob = False):
        # only work with 1 neuron for now
        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        assert len(self.nodes) <= 1
        if N:
            pre_syn = self.gen_traversal_in(['SendsTo', synapse_classes, {'NHP' if high_prob else 'N':(rel,N)}], min_depth=1)
        else:
            if high_prob:
                pre_syn = self.gen_traversal_in(['SendsTo', synapse_classes, {'NHP':(rel,0)}], min_depth=1)
            else:
                pre_syn = self.gen_traversal_in(['SendsTo', synapse_classes], min_depth=1)
        if not len(pre_syn.nodes):
            return []
        pre_neuron_list = []
        synapse_rid_to_N = {s._id: s.NHP if high_prob else s.N for s in pre_syn.get_as('obj', edges = False)[0]}
        synapse_rids = ','.join(self._records_to_list(pre_syn.nodes))
        n_rec=self._graph.client.command("""SELECT $path from (traverse in('SendsTo') FROM [{}] maxdepth 1)""".format(synapse_rids))
        ntos = {n[1]:n[0] for n in [re.findall('\#\d+\:\d+', x.oRecordData['$path']) for x in n_rec] if len(n)==2}
        neuron_rids = list(set(ntos.keys()))
        neurons = self.from_rids(self._graph, *neuron_rids)
        neuron_rid_to_obj = {n._id: n for n in neurons.get_as('obj', edges = False)[0]}

        for neu_id, syn_id in ntos.items():
            pre_neuron_list.append((neuron_rid_to_obj[neu_id], synapse_rid_to_N[syn_id]))
        return pre_neuron_list

    @class_method_timer
    def post_synaptic_neurons_with_synapse_count(self, N=None, rel='>', include_inferred=True, include_fragments = False, high_prob = False):
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = {n._id: n for n in self.node_objs}
        links = self._graph.client.command(
            """MATCH {{class: {neuron_class}, where: (@rid in [{rid}]), as: pre}} -SendsTo-> {{class: Synapse, {where} as: syn}} -SendsTo-> {{class: {neuron_class}, as: target}} return pre, syn.{NHP} as N, target""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ))
        if include_inferred:
            links.extend(self._graph.client.command(
            """MATCH {{class: {neuron_class}, where: (@rid in [{rid}]), as: pre}} -SendsTo-> {{class: InferredSynapse, {where} as: syn}} -SendsTo-> {{class: {neuron_class}, as: target}} return pre, syn.{NHP} as N, target""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            )))
        out = {k: {} for k in rids}
        posts = set()
        for link in links:
            pre = get_hash(link.oRecordData['pre'])
            Nsyn = link.oRecordData['N']
            post = get_hash(link.oRecordData['target'])
            out[pre][post] = Nsyn
            posts.add(post)
        post_neurons = {n._id: n for n in self.from_rids(self._graph, *list(posts)).node_objs}

        return [(rids[pre], post_neurons[post], N) for pre, v in out.items() for post, N in v.items()]
        
    @class_method_timer
    def _post_synaptic_neurons_with_synapse_count(self, N=None, rel='>', include_inferred=True, high_prob = False):
        # only work with 1 neuron for now
        assert len(self.nodes) <= 1
        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        if N:
            post_syn = self.gen_traversal_out(['SendsTo', synapse_classes, {'NHP' if high_prob else 'N':(rel,N)}], min_depth=1)
        else:
            if high_prob:
                post_syn = self.gen_traversal_out(['SendsTo', synapse_classes, {'NHP':(rel,0)}], min_depth=1)
            else:
                post_syn = self.gen_traversal_out(['SendsTo', synapse_classes], min_depth=1)
        if not len(post_syn.nodes):
            return []
        
        post_neuron_list = []
        synapse_rid_to_N = {s._id: s.NHP if high_prob else s.N for s in post_syn.node_objs}
        synapse_rids = ','.join(self._records_to_list(post_syn.nodes))
        n_rec=self._graph.client.command("""SELECT $path from (traverse out('SendsTo') FROM [{}] maxdepth 1) where $depth=1""".format(synapse_rids))
        ntos = {n[1]:n[0] for n in [re.findall('\#\d+\:\d+', x.oRecordData['$path']) for x in n_rec] if len(n)==2}
        neuron_rids = list(set(ntos.keys()))
        neurons = self.from_rids(self._graph, *neuron_rids)
        neuron_rid_to_obj = {n._id: n for n in neurons.node_objs}

        for neu_id, syn_id in ntos.items():
            post_neuron_list.append((neuron_rid_to_obj[neu_id], synapse_rid_to_N[syn_id]))
        return post_neuron_list

    @class_method_timer
    def get_connecting_synapses(self, N=None, rel='>', include_inferred = True, include_fragments = False, high_prob = False):
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = self.node_rids
        cc = self.__class__(self._graph, QueryString(
            """select expand(syn) from (MATCH {{class: {neuron_classes}, where: (@rid in [{rid}])}} -SendsTo-> {{class: Synapse, {where} as: syn}} -SendsTo-> {{class: {neuron_classes}, where: (@rid in [{rid}])}} return syn)""".format(
                rid = ','.join(rids), neuron_classes = neuron_classes,
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ), 'sql'), debug = self.debug)
        if include_inferred:
            cc = cc + self.__class__(self._graph, QueryString(
                """select expand(syn) from (MATCH {{class: {neuron_classes}, where: (@rid in [{rid}])}} -SendsTo-> {{class: InferredSynapse, {where} as: syn}} -SendsTo-> {{class: {neuron_classes}, where: (@rid in [{rid}])}} return syn)""".format(
                    rid = ','.join(rids),  neuron_classes = neuron_classes,
                    where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                    NHP = 'NHP' if high_prob else 'N'
                ), 'sql'), debug = self.debug)
        return cc

    @class_method_timer
    def _get_connecting_synapses_old(self, N=None, rel='>', include_inferred = True):
        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        if N:
            return self.gen_traversal_out(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1)
        else:
            return self.gen_traversal_out(['SendsTo', synapse_classes], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes], min_depth=1)

    @class_method_timer
    def add_connecting_synapses(self, N=None, rel='>', include_inferred = True, include_fragments = False, high_prob = False):
        return self + self.get_connecting_synapses(N = N, rel = rel, include_inferred = include_inferred, 
                include_fragments = include_fragments, high_prob = high_prob)


    @class_method_timer
    def _add_connecting_synapses_old(self, N=None, rel='>',include_inferred=True):
        if N:
            return self + (self.gen_traversal_out(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1))
        else:
            return self + (self.gen_traversal_out(['SendsTo', synapse_classes], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes], min_depth=1))

    @class_method_timer
    def add_connecting_synapses1(self, N=None, rel='>',include_inferred = True, include_fragments = False, high_prob = False):
        neuron_classes = 'NeuronAndFragment' if include_fragments else 'Neuron'
        rids = {n._id: n for n in self.node_objs}
        links = self._graph.client.command(
            """MATCH {{class: Neuron, where: (@rid in [{rid}])}} -SendsTo-> {{class: Synapse, {where} as: syn}} -SendsTo-> {{class: Neuron, where: (@rid in [{rid}])}} return syn""".format(
                rid = ','.join(list(rids.keys())), neuron_class = neuron_classes, 
                where = 'where: ({} {} {}),'.format('NHP' if high_prob else 'N', rel, 0 if N is None else N) if N is not None else 'where: (NHP {} {}),'.format(rel, 0 if N is None else N) if high_prob else '',
                NHP = 'NHP' if high_prob else 'N'
            ))

        synapse_classes = ['Synapse', 'InferredSynapse'] if include_inferred else 'Synapse'
        if N:
            return self + (self.gen_traversal_out(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes, {'N':(rel,N)}], min_depth=1))
        else:
            return self + (self.gen_traversal_out(['SendsTo', synapse_classes], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', synapse_classes], min_depth=1))

    def get_connecting_synapsemodels(self):
        return self.gen_traversal_out(['SendsTo', 'SynapseModel','instanceof'], min_depth=1) & \
                self.gen_traversal_in(['SendsTo', 'SynapseModel','instanceof'], min_depth=1)

    def get_connected_ports(self):
        return self.gen_traversal_out(['SendsTo', 'Port']) + self.gen_traversal_in(['SendsTo', 'Port'])

    @class_method_timer
    def gen_traversal_in(self, *args, **kwargs):
        return self._gen_traversal('in', args, kwargs)

    @class_method_timer
    def gen_traversal_out(self, *args, **kwargs):
        return self._gen_traversal('out', args, kwargs)

    def _gen_traversal(self, direction, args, kwargs):
        '''
            Args must be tuples or list of [edge_types, cls (optional), instanceof_or_cls (optional),
                                            Dict containing parameters to filter the nodes for this particular stage of traversal (optional)]
                  or strings of only edge_types
        '''
        assert len(args)>0
        class_list = list(self._graph.registry.keys())
        rid_list = self.node_rids
        if len(rid_list) == 0:
            return self.empty_query(self._graph, debug = self.debug)
            # return self.__class__(self._graph, QueryString("""select from DataSource where name = "uiyth" ""","sql"),
            #                       debug = self.debug, edges = self.edges)
        q = dict()
        dq = {}
        q['$q0'] = "$q0 = (select from [%s])" % ", ".join(rid_list)
        #dq['$q0'] = "$q0 = (select from (%s))" % self._disp_query

        if 'min_depth' in kwargs:
            assert isinstance(kwargs['min_depth'], numbers.Integral)
            min_depth = kwargs['min_depth']
        else:
            min_depth = 0

        if 'max_depth' in kwargs:
            assert isinstance(kwargs['max_depth'], numbers.Integral)
            max_depth = kwargs['max_depth']
        else:
            max_depth = len(args) + 1

        for t, a in enumerate(args):
            a = _list_repr(a)
            assert len(a) in (1, 2, 3, 4), \
                "Args must be tuples or list of [edge_types, cls (optional), instanceof_or_cls (optional)," +  \
                "Dict containing parameters to filter the nodes for this particular stage of traversal (optional)], or strings of only edge_types"

            arg_dict = {}
            if len(a)==3:
                for v in a[0:min(len(a), 2)]:
                    if isinstance(v, list) or isinstance(v, tuple):
                        assert all(vv in class_list for vv in v), 'Invalid Relationship or Node class'
                    else:
                        assert(v in class_list), 'Invalid Relationship or Node class'
                if a[2]=='instanceof':
                    arg_dict['instanceof'] = a[1]
                elif a[2]=='cls':
                    arg_dict['cls'] = a[1]
                else:
                    arg_dict['cls'] = a[1]
                    if isinstance(a[2],dict): arg_dict.update(a[2])
            elif len(a)==4:
                for v in a[0:min(len(a), 2)]:
                    if isinstance(v, list) or isinstance(v, tuple):
                        assert all(vv in class_list for vv in v), 'Invalid Relationship or Node class'
                    else:
                        assert(v in class_list), 'Invalid Relationship or Node class'
                if a[2]=='instanceof':
                    arg_dict['instanceof'] = a[1]
                else:
                    arg_dict['cls'] = a[1]
                if isinstance(a[3],dict): arg_dict.update(a[3])
            elif len(a) == 2:
                if isinstance(a[1], dict):
                    assert all(v in class_list for v in a[0:min(len(a),1)]), 'Invalid Relationship class'
                    arg_dict.update(a[1])
                else:
                    for v in a[0:min(len(a), 2)]:
                        if isinstance(v, list) or isinstance(v, tuple):
                            assert all(vv in class_list for vv in v), 'Invalid Relationship or Node class'
                        else:
                            assert(v in class_list), 'Invalid Relationship or Node class'
                    arg_dict['cls'] = a[1]

            classes, attrs, depth, columns = _kwargs(arg_dict)

            attrs_query = ""
            if attrs and classes:
                attrs_query = " and (" + " and ".join(attrs) + ") "
            elif (not classes) and attrs:
                attrs_query = " where (" + " and ".join(attrs) + ") "

            #relationships = [direction + "('%s')" % (a if isinstance(a, str) else a[0]) for a in args[:t+1]]
            relationships = [direction + "('%s')" % (x if isinstance(x, str) else x[0]) for x in args[t:t+1]]

            var = '$q' + str(t+1)

            #q[var] = "%s = (select from (select expand(%s) from [%s]) %s)" % \
            #            (var, ".".join(relationships), ", ".join(rid_list), classes)
            q[var] = "%s = (select from (select expand(%s) from %s) %s %s)" % \
                        (var, ".".join(relationships), '$q' + str(t), classes, attrs_query)

            #dq[var] = "%s = (select from (select expand(%s) from (%s)) %s)" % \
            #            (var, ".".join(relationships),self._disp_query, classes)

        # _, _, _, columns = _kwargs(kwargs) #wrong
        query = """ select expand($q) let %s, $q = unionall(%s) """ % \
                (", ".join(list(q.values())), ",".join(list(q.keys())[min_depth:max_depth]))
        disp_query = ""
        #disp_query = "select expand($q) let %s, $q = unionall(%s) " % \
        #        (", ".join(dq.values()[min_depth:max_depth]), ",".join(dq.keys()[min_depth:max_depth]))

        return self.__class__(self._graph, QueryString(query, "sql"), debug = self.debug)

    def path_to_neurons(self, unames, synapse_threshold = 10, max_hops = 2):
        q1 = self.__class__(self._graph, QueryString("""select from Neuron where uname in ["{}"]""".format('","'.join(unames)), "sql"), debug = self.debug)
        return self.path_to(q1)[0]

    def direct_path_to(self, q, synapse_threshold = 10, max_hops = 2):
        return self.path_to(q, synapse_threshold = synapse_threshold,
                            max_hops = max_hops, exclude_path = True)

    def path_to(self, q, synapse_threshold = 10, max_hops = 2,
                exclude_path = True, timeout = 120):
        if max_hops == 0:
            return self.synapses_to(q, synapse_threshold = synapse_threshold)

        from_neurons = self._records_to_list(self.has(cls = 'Neuron').nodes)
        to_neurons = self._records_to_list(q.has(cls = 'Neuron').nodes)
        exclude_rids = set(from_neurons + to_neurons)
        path_rids = set()
        all_paths = []
        used_time = 0
        try:
            for hop in range(1, max_hops+1):
                template = """
                -SendsTo->
                {{class: Neuron, as: hop{hop}, where: (@rid not in [{exclude_rids}])}}
                -SendsTo->
                {{class: Synapse, where: (N >= {synapse_threshold})}}
                """

                query = """select from(
                MATCH
                {{class: Neuron, as: source, where: (@rid in [{from_rids}]) }}
                -SendsTo->
                {{class: Synapse, where: (N >= {synapse_threshold})}}
                """.format(from_rids = ','.join(from_neurons),
                           synapse_threshold = synapse_threshold) \
                + \
                ''.join([
                template.format(hop = i+1,
                                synapse_threshold = synapse_threshold,
                                exclude_rids = ','.join(exclude_rids)) for i in range(hop)
                ]) \
                + \
                """
                -SendsTo->
                {{class: Neuron, as: dest, where: (@rid in [{to_rids}]) }}
                return source, {hops}, dest)
                timeout {timeout}""".format(
                    to_rids = ','.join(to_neurons),
                    hops = ','.join(['hop{}'.format(i+1) for i in range(hop)]),
                    timeout = int((timeout-used_time)*1000))

                start = time.time()
                q =  self._graph.client.command(query)
                used_time += time.time()-start

                paths = [[get_hash(node.oRecordData['source'])] + \
                         [get_hash(node.oRecordData['hop{}'.format(i+1)]) for i in range(hop)] + \
                         [get_hash(node.oRecordData['dest'])] for node in q]
                all_paths.extend(paths)
                #sum([[node.oRecordData['hop{}'.format(i+1)].get_hash() for node in q] for i in range(hop)], [])
                path_rids.update(set(sum([path[1:-1] for path in paths], [])))
                if exclude_path:
                    exclude_rids.update(path_rids)
        except PyOrientCommandException:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            # pdb.set_trace()
            if 'timeout' in exc_value.errors[0]:
                if hop == 1:
                    warn('Execution timeout, no results returned')
                else:
                    warn('Execution timeout, returnning results for maximum of {} hop(s)'.format(hop-1),
                         RuntimeWarning)
            else:
                raise
        return self.from_rids(self._graph, *list(path_rids)), all_paths


    def path_to2(self, q, synapse_threshold = 10, max_hops = 2,
                 exclude_path = True, timeout = 30):
        if max_hops == 0:
            return self.synapses_to(q, synapse_threshold = synapse_threshold)
        from_neurons = self._records_to_list(self.has(cls = 'Neuron').nodes)
        to_neurons = self._records_to_list(q.has(cls = 'Neuron').nodes)
        query = """
        select distince(pp) from
        (MATCH
        {{class: Neuron, where: (@rid in [{from_rids}]) }}
        -SendsTo->{{}}-SendsTo->
        {{ maxdepth: {nhop}, where: ($depth%2 = 1 and @rid not in [{exclude_rids}]), pathAlias: pp}}
        -SendsTo->{{}}-SendsTo->
        {{class: Neuron, where: (@rid in [{to_rids}]) }}
        return pp) timeout {timeout}""".format(from_rids = ','.join(from_neurons),
                   exclude_rids = ','.join(exclude_rids),
                   to_rids = ','.join(to_neurons),
                   nhop = max_hops+1,
                   timeout = timeout * 1000)


    def synapses_to(self, q, synapse_threshold = 10):
        from_neurons = self._records_to_list(self.has(cls = 'Neuron').nodes)
        to_neurons = self._records_to_list(q.has(cls = 'Neuron').nodes)
        q = self._graph.client.command(
        """
        MATCH
        {{class: Neuron, where: (@rid in [{from_rids}]) }}
        -SendsTo->
        {{class: Synapse, as: synapses, where: (N >= {synapse_threshold})}}
        -SendsTo->
        {{class: Neuron, where: (@rid in [{to_rids}]) }}
        return synapses
        """.format(from_rids = ','.join(from_neurons),
                   to_rids = ','.join(to_neurons),
                   synapse_threshold = synapse_threshold)
        )
        # synapse_rids = []
        # for node in q:
        #     data = node.oRecordData['synapses']
        #     if isinstance(data, tuple):
        #         if data[0] == 'OrientRecordLink':
        #             rid = '#{}:{}'.format(data[1], data[2])
        #             synapse_rids.append(rid)
        #     elif isinstance(data, OrientRecordLink):
        #         synapse_rids.append(data.get_hash())
        # synapse_rids = list(set(synapse_rids))
        synapse_rids = list(set([get_hash(node.oRecordData['synapses']) for node in q]))
        return self.from_rids(self._graph, *synapse_rids)

    def export_graph(self, graph_name, as_type='df', stored_as='gpickle', compression=''):
        g = self.get_as(as_type)
        if as_type=='df':
            g[0].to_csv(graph_name+'_nodes.csv', index=True)
            g[1].to_csv(graph_name+'_edges.csv', index=False)
            return 'Saved'
        elif as_type=='nx':
            if stored_as=='gpickle':
                nx.nx.write_gpickle(g, graph_name+'.gpickle'+compression)
                return 'Saved'
            elif stored_as=='gexf':
                nx.nx.write_gexf(g, graph_name+'.gexf'+compression)
                return 'Saved'
            else:
                raise UnsupportedView(stored_as)
        raise UnsupportedView(as_type)

    def import_graph(self, graph_name, as_type='df', stored_as='csv', compression=''):
        client = self._graph.client
        if as_type=='df':
            df_node = pd.pd.read_csv(graph_name+'_nodes.csv', index_col=0)
            df_edge = pd.pd.read_csv(graph_name+'_edges.csv', index_col=False)
            pd.pandas_to_orient(client, df_node, df_edge)
            return True
        elif as_type=='nx':
            if stored_as=='gpickle':
                g = nx.nx.read_gpickle(graph_name+'.gpickle'+compression)
                nx.nx_to_orient(client, g)
                return True
            elif stored_as=='gexf':
                g = nx.nx.read_gexf(graph_name+'.gexf'+compression)
                nx.nx_to_orient(client, g)
                return True
        return False

    @class_method_timer
    def __or__(self, other):
        assert isinstance(other, self.__class__)
        return self.__class__(self._graph, ('|', self._query, other._query),
                              init_nodes=self._dict_union(self._nodes, other._nodes),
                              executed = True,
                              debug = (self.debug | other.debug))#'(%s\n|\n%s)' %(self._disp_query, other._disp_query))

    __add__ = __or__

    @class_method_timer
    def __sub__(self, other):
        assert isinstance(other, self.__class__)
        return self.__class__(self._graph, ('-', self._query, other._query),
                              init_nodes=self._dict_difference(self._nodes, other._nodes),
                              executed = True,
                              debug = (self.debug | other.debug))#'(%s\n-\n%s)' %(self._disp_query, other._disp_query))

    @class_method_timer
    def __and__(self, other):
        assert isinstance(other, self.__class__)
        return self.__class__(self._graph, ('&', self._query, other._query),
                              init_nodes=self._dict_intersection(self._nodes, other._nodes),
                              executed = True,
                              debug = (self.debug | other.debug))#'(%s\n&\n%s)' %(self._disp_query, other._disp_query))

    __mul__ = __and__

    @class_method_timer
    def __xor__(self, other):
        assert isinstance(other, self.__class__)
        return self.__class__(self._graph, ('^', self._query, other._query),
                              init_nodes=self._dict_symmetric_difference(self._nodes, other._nodes),
                              executed = True,
                              debug = (self.debug | other.debug))#'(%s\n^\n%s)' %(self._disp_query, other._disp_query))

    @class_method_timer
    def __ior__(self, other):
        assert isinstance(other, self.__class__)
        self._query = ('|', self._query, other._query)
        self._nodes = self._dict_union(self._nodes, other._nodes)
        #self._disp_query = '(%s\n|\n%s)' %(self._disp_query, other._disp_query)

    __iadd__ = __ior__

    @class_method_timer
    def __isub__(self, other):
        assert isinstance(other, self.__class__)
        self._query = ('-', self._query, other._query)
        self._nodes = self._dict_difference(self._nodes, other._nodes)
        #self._disp_query = '(%s\n-\n%s)' %(self._disp_query, other._disp_query)

    @class_method_timer
    def __iand__(self, other):
        assert isinstance(other, self.__class__)
        self._query = ('&', self._query, other._query)
        self._nodes = self._dict_intersection(self._nodes, other._nodes)
        #self._disp_query = '(%s\n&\n%s)' %(self._disp_query, other._disp_query)

    @class_method_timer
    def __ixor__(self, other):
        assert isinstance(other, self.__class__)
        self._query = ('^', self._query, other._query)
        self._nodes = self._dict_symmetric_difference(self._nodes, other._nodes)
        #self._disp_query = '(%s\n^\n%s)' %(self._disp_query, other._disp_query)

    @class_method_timer
    def __eq__(self, other):
        """
        Check whether nodes returned by queries are equivalent.
        """

        assert isinstance(other, self.__class__)

        # Queries can only be checked for equality after evaluation:
        if not self._executed:
            return False
        if set(self._nodes) == set(other._nodes):
            return True
        else:
            return False

def _q_repr(attr):
    if isinstance(attr, str):
        return "'" + str(attr) + "'"
    else:
        return str(attr)

def _list_repr(attr):
    if not(isinstance(attr, list)):
        if isinstance(attr, (str, numbers.Number)):
            return [attr]
        else:
            return list(attr)
    return attr


def _kwargs(kwargs):
    if 'max_levels' in kwargs:
        assert isinstance(kwargs['max_levels'], numbers.Integral) and kwargs['max_levels'] >= 0

    assert not (('cls' in kwargs) and ('instanceof' in kwargs)), "can't use both cls and instanceof"

    attrs = []
    classes = ""
    columns = ""
    depth = ""

    for k, v in kwargs.items():
        if k=='max_levels':
            depth="while $depth <= %s" %v
        elif k=='instanceof':
            classes = "where @this instanceof '%s'" % v
        else:
            v = _list_repr(v)
            if k=='cls':
                if kwargs[k]:
                    classes = "where @class in %s" % v
            elif k=='cols':
                columns = ", ".join(v)
            elif k=='rid':
                attrs.append("@rid in %s" % v.__repr__().replace("'",""))
            else:
                # if isinstance(k, (tuple, list)) and len(k) > 1:
                #     if len(v) == 1 and isinstance(v[0],(str,bytes)) and len(v[0])>=2 and v[0][:2] == '/r':
                #         attrs.append(""" any(%s) matches "%s" """ % (','.join(k), v[0][2:]))
                #     elif (len(v) ==2 and isinstance(v[0],(str,bytes)) and len(v[0])
                #     and v[0] in ['<','>','=','<=','>=']):
                #         attrs.append("any(%s) %s %s" % (','.join(k),v[0],v[1]))
                #     else:
                #         attrs.append("any(%s) in %s" % (','.join(k), v))
                # else:
                if len(v) == 1 and isinstance(v[0],(str,bytes)) and len(v[0])>=2 and v[0][:2] == '/r':
                    # regex match
                    attrs.append("""%s matches "%s" """ % (k, v[0][2:]))
                elif (len(v) ==2 and isinstance(v[0],(str,bytes)) and len(v[0])
                and v[0] in ['<','>','=','<=','>=']):
                    attrs.append("%s %s %s" % (k,v[0],v[1]))
                else:
                    attrs.append("%s in %s" % (k, v))
    return classes, attrs, depth, columns


if __name__ == '__main__':
    # from networkx import *
    from pyorient.ogm import Config, Graph
    import numpy as np
    from neuroarch.utils import is_rid, iterable
    from neuroarch.query import *
    from neuroarch.diff import diff_nodes, diff_edges
    from neuroarch.apply_diff import apply_node_diff, apply_edge_diff
    import pandas as pd

    from neuroarch.models import Node, Relationship

    from pyorient.ogm.declarative import declarative_node, declarative_relationship

    SchemaNode = declarative_node()
    SchemaRelationship = declarative_relationship()
    graph = Graph(Config.from_url('/na_server', 'root', 'root', initial_drop=False))

    classes_from_schema = graph.build_mapping(SchemaNode, SchemaRelationship, auto_plural=True)

    graph.include(Node.registry)
    graph.include(Relationship.registry)
    print(graph, type(graph))

    q2=QueryWrapper(graph,QueryString("select from CartridgeModel where name in ['cart114']",'sql'))
    print(q2)
    n1 = q2.traverse_owns(max_levels=3) #, cls = 'MorrisLecar')
    n2 = n1.get_as()[0]

    threshold = list(np.arange(0,1.,0.1)) + [np.nan]*(len(n2)-len(np.arange(0,1.,0.1)))
    model = ['a']*(len(n2)-4) + [np.nan]*4
    n2['threshold'] = threshold
    a={'#111:1':{'name': 'a', 'n_outputs': 2, 'class': 'MorrisLecar'}}
    b={'#112:1':{'name': 'b', 'n_outputs': 2, 'class': 'MorrisLecar'}}
    n2 = n2.ix[:]
    n2 = n2.append(pd.DataFrame(a).T)
    n2 = n2.append(pd.DataFrame(b).T)
    n2.drop(['V2'], inplace=True, axis=1)
    q3 = QueryWrapper(graph,QueryString("select from Neuron limit 20",'sql'))



#!/usr/bin/env python

"""
Pandas/NetworkX graph conversion utilities.
"""

# Copyright (c) 2015, Lev Givon
# All rights reserved.
# Distributed under the terms of the BSD license:
# http://www.opensource.org/licenses/bsd-license

import pandas as pd
import networkx as nx
import numpy as np

def _find_field_types(d):
    """
    Create @fieldTypes property value for setting types when creating nodes with JSON content.    
    """

    field_types = []
    for k, v in d.items():
        if isinstance(v, np.float32):
            field_types.append('%s=f' % k)
        elif isinstance(v, float) or isinstance(v, np.float64):
            field_types.append('%s=d' % k)
    return ','.join(field_types)

def nx_to_pandas(g):
    """
    Convert directed multigraph stored as NetworkX MultiDiGraph into Pandas DataFrame.

    Parameters
    ----------
    g : networkx.MultiDiGraph
        Directed multigraph to convert.

    Returns
    -------
    df_node : pandas.DataFrame
        Node attributes. The NetworkX identifiers are used as the DataFrame
        index.
    df_edge : pandas.DataFrame
        Edge attributes. Source and destination node identifiers are stored in
        `out` and `in` columns, respectively.
    """

    assert isinstance(g, nx.MultiDiGraph)
    node_data = g.nodes(data=True)
    index = [n[0] for n in node_data]
    props = [n[1] for n in node_data]
    df_node = pd.DataFrame.from_records(props, index)

    index = range(g.number_of_edges())
    props = []
    for e in g.edges(data=True):
        p = {'out': e[0], 'in': e[1]}
        p.update(e[2])
        props.append(p)
    df_edge = pd.DataFrame.from_records(props, index)
    return df_node, df_edge

def pandas_to_nx(df_node, df_edge):
    """
    Convert directed multigraph stored as Pandas DataFrames into NetworkX MultiDiGraph.

    Parameters
    ----------
    df_node : pandas.DataFrame
        Node attributes. The index contents are assumed to be the node identifiers.
    df_edge : pandas.DataFrame
        Edge attributes. Source and destination node identifiers are stored in
        `out` and `in` columns, respectively.

    Returns
    -------
    g : networkx.MultiDiGraph
        Directed multigraph to convert.
    """

    assert isinstance(df_node, pd.DataFrame)
    assert isinstance(df_edge, pd.DataFrame)
    g = nx.MultiDiGraph()
    for id, props in zip(df_node.index, df_node.to_dict('record')):
        g.add_node(id, props)
    for id, props in zip(df_edge.index, df_edge.to_dict('record')):
        from_id = props['out']
        to_id = props['in']
        del props['out']
        del props['in']
        g.add_edge(from_id, to_id, **props)
    return g
